import com.javageeks.classloader.ClassLoaderStrategy;
import com.javageeks.classloader.StrategyClassLoader;
import DE.fub.inf.JVM.ClassGen.*;
import DE.fub.inf.JVM.Constants;

/**
 * DynamicGen uses the ClassFile API to build a Java class at
 * runtime. Obviously, this class depends on the ClassFile API,
 * found in the JavaClass.jar file in the JavaClass directory
 * (under this one). Either place the JavaClass.jar file in your
 * Extensions directory (by default, in the jre\lib\ext directory
 * under your JDK or JRE installation), or else add this directory
 * plus "JavaClass" to the CLASSPATH. Alternatively, move the "DE"
 * and "gnu" subdirectories into this directory, and it will run
 * without necessary modification to your environment (that is, the
 * classes will get picked up by the AppClassLoader/ExtClassLoader
 * "primordial" ClassLoader scheme in JDK 1.2).
 */
public class DynamicGen
{
    /**
     * Everything happens in main(). We first build an anonymous
     * ClassLoaderStrategy to dynamically build the HelloWorld class. Please
     * note that the actual HelloWorld-construction code comes from the
     * ClassFile API demonstration code. Once the ClassLoaderStrategy
     * is built, we put it into a StrategyClassLoader, attempt to construct
     * an instance of HelloWorld, and call its main() method through
     * Reflection (since the class doesn't exist at compile-time, there's no
     * way we can cast the Object returned to HelloWorld; such is the price
     * we pay for this flexibility).
     */
    public static void main (String args[])
        throws Exception
    {
        ClassLoaderStrategy cls = new ClassLoaderStrategy()
        {
            public byte[] findClassBytes(String nm)
            {
                // Check to make sure HelloWorld is the only one requested;
                // we don't know how to build anything else!
                if (!"HelloWorld".equals(nm))
                {
                    return null;
                }

                ClassGen cg = new ClassGen("HelloWorld", "java.lang.Object",
                          "<generated>", Constants.ACC_PUBLIC | Constants.ACC_SUPER,
                          null);
                          
                ConstantPoolGen cp = cg.getConstantPool();
                
                InstructionList il = new InstructionList();
                
                MethodGen mg =
                    new MethodGen(Constants.ACC_STATIC | Constants.ACC_PUBLIC,
                    Type.VOID,              // return type
                    new Type[]
                    {
                        new ArrayType(Type.STRING, 1) 
                    },
                    new String[] { "argv" }, // arg names
                    "main", "HelloWorld",    // method, class
                    il, cp);

                // Add often needed constants to constant pool.
                //
                int br_index   = cp.addClass("java.io.BufferedReader");
                int ir_index   = cp.addClass("java.io.InputStreamReader");
                int system_out = cp.addFieldref("java.lang.System", "out", // System.out
                                "Ljava/io/PrintStream;");
                int system_in  = cp.addFieldref("java.lang.System", "in",  // System.in
                                "Ljava/io/InputStream;");

                // Create BufferedReader object and store it in local variable `in'.
                //
                il.append(new NEW(br_index));
                il.append(new DUP());
                il.append(new NEW(ir_index));
                il.append(new DUP());
                il.append(new GETSTATIC(system_in));

                // Call constructors, i.e. BufferedReader(InputStreamReader())
                //
                il.append(new INVOKESPECIAL(cp.addMethodref("java.io.InputStreamReader",
                    "<init>", "(Ljava/io/InputStream;)V")));
                il.append(new INVOKESPECIAL(cp.addMethodref("java.io.BufferedReader",
                    "<init>", "(Ljava/io/Reader;)V")));

                // Create local variable `in'
                //
                LocalVariableGen lg = mg.addLocalVariable("in",
                    new ObjectType("java.io.BufferedReader"), null, null);
                int in = lg.getIndex();
                lg.setStart(il.append(new ASTORE(in))); // `i' valid from here

                // Create local variable `name'
                //
                lg = mg.addLocalVariable("name", Type.STRING, null, null);
                int name = lg.getIndex();
                il.append(new ACONST_NULL());
                lg.setStart(il.append(new ASTORE(name))); // `name' valid from here

                // try
                // {
                //
                InstructionHandle try_start = il.append(new GETSTATIC(system_out));
                il.append(new PUSH(cp, "Please enter your name> "));
                il.append(new INVOKEVIRTUAL(cp.addMethodref("java.io.PrintStream",
                    "print", "(Ljava/lang/String;)V")));
                il.append(new ALOAD(in));
                il.append(new INVOKEVIRTUAL(cp.addMethodref("java.io.BufferedReader",
                    "readLine", "()Ljava/lang/String;")));
                il.append(new ASTORE(name));

                // Upon normal execution we jump behind exception handler, 
                // the target address is not known yet.
                //
                GOTO g = new GOTO(null);
                InstructionHandle try_end = il.append(g);

                // } catch() { ... }
                // Add exception handler: simply return from method
                //
                InstructionHandle handler = il.append(new RETURN());
                mg.addExceptionHandler(try_start, try_end, handler, "java.io.IOException");

                // Normal code continues, now we can set the branch target of the GOTO.
                //
                InstructionHandle ih = il.append(new GETSTATIC(system_out));
                g.setTarget(ih);

                // String concatenation compiles to StringBuffer operations.
                //
                il.append(new NEW(cp.addClass("java.lang.StringBuffer")));
                il.append(new DUP());
                il.append(new PUSH(cp, "Hello, "));
                il.append(new INVOKESPECIAL(cp.addMethodref("java.lang.StringBuffer",
                    "<init>", "(Ljava/lang/String;)V")));
                il.append(new ALOAD(name));
                
                // One can also abstract from using the ugly signature syntax by using
                // the getMethodSignature() method. For example:
                //
                String sig = Type.getMethodSignature(Type.STRINGBUFFER,
                    new Type[] { Type.STRING });
                il.append(new INVOKEVIRTUAL(cp.addMethodref("java.lang.StringBuffer",
                    "append", sig)));

                il.append(new INVOKEVIRTUAL(cp.addMethodref("java.lang.StringBuffer",
                    "toString", "()Ljava/lang/String;")));
                
                il.append(new INVOKEVIRTUAL(cp.addMethodref("java.io.PrintStream",
                    "println", "(Ljava/lang/String;)V")));
                il.append(new RETURN());

                mg.setMaxStack(5); // Needed stack size
                cg.addMethod(mg.getMethod());

                // Add public <init> method, i.e. empty constructor
                //
                cg.addEmptyConstructor(Constants.ACC_PUBLIC);

                // Get JavaClass object and dump it to file.
                //
                return cg.getJavaClass().getBytes();
            }

            // Other ClassLoaderStrategy methods
            public String findLibraryPath(String str) { return null; }
            public java.net.URL findResourceURL(String name) { return null; }
            public java.util.Enumeration findResourcesEnum(String name) { return null; }
        };

        StrategyClassLoader scl = new StrategyClassLoader(cls);

        // Construct na instance of "HelloWorld"
        Class clazz = scl.loadClass("HelloWorld");
        Object obj = clazz.newInstance();
        
        // Now that we have it, we use Reflection to find the main() method and
        // invoke() it; this is the only way to avoid having to have the HelloWorld
        // class around at compile time for type-checking
        java.lang.reflect.Method[] methods = clazz.getMethods();
        for (int i=0; i<methods.length; i++)
        {
            System.out.println("Checking name of " + clazz.getName()
                + "." + methods[i].getName());
            if (methods[i].getName().equals("main"))
            {
                if (methods[i].getReturnType().equals(java.lang.Void.TYPE) &&
                    methods[i].getParameterTypes().length == 1)
                {
                    // methods[i] is the Method that corresponds to the
                    // method "void main(String[])". Call it.
                    //
                    Object ret = methods[i].invoke(obj, new Object[] 
                    {
                        null
                    });
                    if (ret != null)
                        System.out.println("??? main()'s not supposed to " +
                            "return me something!");
                    else
                        break;
                }
            }
        }
    }
}
