import Employee.*;

/**
 * This class provides a regression test and behavior/specification
 * tests for the Employee system.
 */
public class EmployeeTest
{
    public static void main(String[] args)
    {
        try
        {
            // Get our model
            //
            IEmployeeModel model = null;
            
            if (args != null)
                for (int i=0; i< args.length; i++)
                {
                    if (args[i].equals("-model"))
                        model = getEmployeeModel(args[++i]);
                    if (args[i].equals("-build"))
                        build(model);
                }

            if (model==null)
            {
                model = getEmployeeModel();
            }
            
            display(model);
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
    }

    public static void build(IEmployeeModel model)
        throws Exception
    {
        // Build the corporation's Department structure
        //
        IDepartment board = 
            model.createDepartment("Board", null);
        board.setDescription("Executive Board");
            
        IDepartment finance = 
            model.createDepartment("Finance", board);
        finance.setDescription("Manages corporate assets");
        IDepartment humanResources = 
            model.createDepartment("HR", finance);
        humanResources.setDescription("Manages employees");
            
        IDepartment engineering = 
            model.createDepartment("Engineering", board);
        engineering.setDescription("Engineers things");
        IDepartment infoTech = 
            model.createDepartment("IT", engineering);
        infoTech.setDescription("Information Technology");
        IDepartment research = 
            model.createDepartment("R&D", engineering);
        research.setDescription("Research & Development");
            
        IDepartment customerService = 
            model.createDepartment("Customer Service", board);
        customerService.setDescription("Services customers");
        IDepartment techSupport = 
            model.createDepartment("Tech Support", customerService);
        techSupport.setDescription("Supports technical stuff");
        IDepartment mktg = 
            model.createDepartment("Marketing", customerService);
        mktg.setDescription("Markets stuff");
        IDepartment sales = 
            model.createDepartment("Sales", customerService);
        sales.setDescription("Sells stuff");
        
        // Add some positions
        //
        IPosition CEO =
            model.createPosition("CEO", "Chief Executive Officer",
                100000, Integer.MAX_VALUE);

        IPosition CFO =
            model.createPosition("CFO", "Chief Financial Officer",
                CEO.getMinSalary(), CEO.getMaxSalary());

        IPosition CTO =
            model.createPosition("CTO", "Chief Technical Officer",
                CEO.getMinSalary(), CEO.getMaxSalary());
        
        IPosition VP =
            model.createPosition("Vice-President", "Vice-President",
                80000, CEO.getMinSalary());
        
        IPosition mgr =
            model.createPosition("Manager", "Clueless, of course",
                70000, VP.getMinSalary());
        
        IPosition marketer =
            model.createPosition("Marketer", 
                "Markets stuff to customers", 0, 45000);
        
        IPosition developer =
            model.createPosition("Software Developer", 
                "Technical underappreciated underpaid genius",
                45000, CEO.getMinSalary()); // yeah, right
        
        IPosition intern =
            model.createPosition("Intern", 
                "Cheap pawns in the corporate game", 0, 0);
        
        // Add some people
        //
        IPerson jerry =
            model.createPerson("Jerry", "", "Seinfeld",
                "357-24-1357");
        jerry.addEMail("Home email", "jerry@seinfeld.com");
        
        IPerson fred =
            model.createPerson("Fred", "", "Flintstone",
                "000-00-0001");
        fred.addAddress("Home address", "1 Stoney Way", "Bedrock",
            "CA", "00000", "USA");
        fred.addEMail("Home email", "fred@bedrock.com");
        
        IPerson barney =
            model.createPerson(
                "Barney", "", "Rubble", "000-01-0000");
        barney.addEMail("Home email", "barney@bedrock.com");

        // Add some employees
        //
        IEmployee salesDude =
            model.createEmployee(
                "Sales", "", "Dude", "666-34-2375", marketer, 40000);
        salesDude.addPhone("Work phone", "800", "BUY-STUF", "x3092");
        
        IEmployee theBoss =
            model.createEmployee(
                "The", "", "Boss", "276-57-2642", CEO, 150000);
        theBoss.addPhone("Work phone", "666", "555-1212", "");
        theBoss.addEMail("Work email", "boss@company.com");
        
        IEmployee i1 =
            model.createEmployee(
                "Nameless", "", "Intern", "123-45-6789", intern, 0);
        
        IEmployee i2 =
            model.createEmployee(
                "Nameless", "Other", "Intern", "123-98-7654", intern, 0);
        
        IEmployee i3 =
            model.createEmployee(
                "Nameless", "Last", "Intern", "987-65-4321", intern, 10);
        
        // Add some managers
        //
        IManager theBossMgr =
            model.createManager(theBoss);
        
        theBossMgr.addEmployee(salesDude);
        theBossMgr.addEmployee(i1);
        theBossMgr.addEmployee(i2);
        theBossMgr.addEmployee(i3);
        
        board.setManager(theBossMgr);
        finance.setManager(theBossMgr);
        humanResources.setManager(theBossMgr);
        engineering.setManager(theBossMgr);
        infoTech.setManager(theBossMgr);
        research.setManager(theBossMgr);
        customerService.setManager(theBossMgr);
        techSupport.setManager(theBossMgr);
        mktg.setManager(theBossMgr);
        sales.setManager(theBossMgr);
            // The Boss is a pretty busy guy; that, or he's a 
            // real control freak....
    }
        

    public static void display(IEmployeeModel model)
        throws Exception
    {
        // Get all Departments in the database
        //
        System.out.println("\n\nAll Departments:");
        IDepartment[] depts = model.findAllDepartments();
        for (int i=0; i<depts.length; i++)
            printDept(depts[i]);
        System.out.println("\nRootDepartment:");
        IDepartment rootDept = model.findRootDepartment();
        printDept(rootDept);
        
        /*
        // Get all Positions in the database
        //
        System.out.println("\n\nAll Positions:");
        IPosition[] positions = model.findAllPositions();
        for (int i=0; i<positions.length; i++)
            printPosition(positions[i]);

        // Get all Persons in the database
        //
        System.out.println("\n\nAll Persons:");
        IPerson[] persons = model.findAllPersons();
        for (int i=0; i<persons.length; i++)
            printPerson(persons[i]);
        
        // Get all Employees in the database
        //
        System.out.println("\n\nAll Employees:");
        IEmployee[] employees = model.findAllEmployees();
        for (int i=0; i<employees.length; i++)
            printEmployee(employees[i]);
        
        
        // Get all Managers in the database
        System.out.println("\n\nAll Managers:");
        IManager[] mgrs = model.findAllManagers();
        for (int i=0; i<mgrs.length; i++)
            printManager(mgrs[i]);
            
        // Show persons with last name of "Smith"
        // (should be empty)
        System.out.println("\n\nFind Persons with lName 'Smith':");
        IPerson[] smiths = model.findPersonsByLastName("Smith");
        if (smiths != null)
            for (int i=0; i<smiths.length; i++)
                printPerson(smiths[i]);
        else
            System.out.println("    (None found)");
            
        // Show persons with last name of "Intern"
        // (should be empty)
        System.out.println("\n\nFind Persons with lName 'Intern':");
        IPerson[] interns = model.findPersonsByLastName("Intern");
        if (interns != null)
            for (int i=0; i<interns.length; i++)
                printPerson(interns[i]);
        else
            System.out.println("    (None found)");
        */
    }
    
    private static IEmployeeModel getEmployeeModel()
    {
        return getEmployeeModel("Hashtable");
    }
    private static IEmployeeModel getEmployeeModel(String whichModel)
    {
        /**** HashtableModel ****/
        if (whichModel.equals("Hashtable"))
        {
            return new HashtableModel();
        }

        /**** RDBMSModel ****/
        if (whichModel.equals("RDBMS"))
        {
            // Set up JDBC Connection
            java.sql.Connection conn = null;
            try
            {
                // Load the JDBC driver into the JVM
                Class.forName("jdbc.idbDriver");
                java.util.Properties p = new java.util.Properties();
                conn = java.sql.DriverManager.getConnection(
                        "jdbc:idb:employee.prp", p);
            }
            catch (Exception ex)
            {
                ex.printStackTrace();
                return null;
            }
        
            // Build the EmployeeModel
            return new RDBMSModel(conn);
        }

        /**** JSDTModel ****/
        if (whichModel.equals("JSDT"))
        {
            return new JSDTModel("Client1", "localhost");
        }

        return null;
    }

    private static void printPerson(IPerson person)
        throws Exception
    {
        System.out.println(person.getLastName() + ", " +
            person.getFirstName() + " " +
            person.getMiddleName() + " (" +
            person.getSSN() + ")");
        
        IContactInfo[] contactInfo = person.getContactInfo();
        for (int j=0; j<contactInfo.length; j++)
        {
            // Print ContactInfo
            System.out.println("    " + 
                contactInfo[j].toString() +
                (contactInfo[j].getDescription() != null ? 
                    " (" +
                    contactInfo[j].getDescription() + 
                    ")" : ""));
        }
    }
    private static void printEmployee(IEmployee employee)
        throws Exception
    {
        System.out.println(employee.getLastName() + ", " +
            employee.getFirstName() + " " +
            employee.getMiddleName() + " (" +
            employee.getSSN() + "): " +
            employee.getPosition().getTitle() + ", making " +
            employee.getSalary());
        
        IContactInfo[] contactInfo = employee.getContactInfo();
        for (int j=0; j<contactInfo.length; j++)
        {
            // Print ContactInfo
            System.out.println("    " + 
                contactInfo[j].toString() +
                (contactInfo[j].getDescription() != null ? 
                    " (" +
                    contactInfo[j].getDescription() + 
                    ")" : ""));
        }
    }
    private static void printDept(IDepartment dept)
        throws Exception
    {
        System.out.println(dept.getTitle() + ": " + 
            dept.getDescription());
        if (dept.getManager() != null)
        {
            System.out.println("    Run by " +
                dept.getManager().getLastName() + ", " +
                dept.getManager().getFirstName() + " " +
                dept.getManager().getMiddleName());
        }
        else
            System.out.println("    (No Manager)");
            
        IDepartment[] childDepts =
            dept.getContainedDepartments();
        if (childDepts != null)
        {
            System.out.print("Contains: ");
            for (int i=0; i<childDepts.length; i++)
                System.out.print(childDepts[i].getTitle() + " ");
            System.out.println("");
        }
    }
    private static void printPosition(IPosition pos)
        throws Exception
    {
        System.out.println(pos.getTitle() + "(" +
            pos.getMinSalary() + " to " +
            pos.getMaxSalary() + "): " + 
            pos.getDescription());
    }
    private static void printManager(IManager mgr)
        throws Exception
    {
        System.out.println(mgr.getLastName() + ", " +
            mgr.getFirstName() + " " +
            mgr.getMiddleName());
        IEmployee[] employees = mgr.getEmployees();
        for (int i=0; i<employees.length; i++)
        {
            System.out.println("    " +
                employees[i].getLastName() + ", " +
                employees[i].getFirstName() + " " +
                employees[i].getMiddleName());
        }
    }
}    
