//=========================================================
// $Id: JVMService.cpp,v 1.2 1999/07/15 10:15:34 tneward Exp $
//=========================================================
// NTService code derived from "Simplest Service", from
// "Win32 System Services: The Heart of Windows NT", by
// Marshall Brain, Prentice-Hall, 1994, ISBN 0-13-097825-6
//=========================================================
// $Log: JVMService.cpp,v $
// Revision 1.2  1999/07/15 10:15:34  tneward
// Periodic update
//
// Revision 1.1  1999/07/09 11:33:26  tneward
// Periodic update
//
//=========================================================

// In order to be loaded by the JVMService, a Java class
// must have the following methods:
//
// pseudo-interface NTService
// {
//     public void start(String[] args);
//     public void stop();
//     public void pause();
//     public void resume();
// }
//
// Because these will be loaded via JNI and called by name,
// no specific interface needs to be implemented. Implementors
// must take care that start/stop/pause/resume are appropriately
// thread-synchronized, as the NT ServiceControlManager will
// use multiple threads to call back into the Service.
//


// OS headers
#include <windows.h>

// ANSI C++ headers
#include <iostream>
#include <string>
#include <vector>
using namespace std;

// JNI headers
#include <jni.h>
#include "java.h"

// ATL headers (for CRegKey)
#include <atlbase.h>
#include <atlimpl.cpp>


// Really cheap trick to avoid having to do multiple-file
// compilations; not recommended for the faint-of-heart
#include "java_md.c"


//=========================================================
// JVM-related constants & globals
//

// Globals
JavaVM* vm = (JavaVM*)0;
JNIEnv* env = (JNIEnv*)0;
jclass serviceClass = 0;
jmethodID startMethod = 0;
jmethodID stopMethod = 0;
jmethodID pauseMethod = 0;
jmethodID resumeMethod = 0;

jboolean debug;

/*
 * List of VM options to be specified when the VM is created.
 */
static JavaVMOption* options;
static int numOptions;
static int maxOptions;


//=========================================================
// NTService-related constants & globals
//

// Constants
char* SERVICE_NAME = "JVMService";

// Globals
HANDLE terminateEvent = NULL;
SERVICE_STATUS_HANDLE serviceStatusHandle;
BOOL paused = FALSE;
BOOL running = FALSE;
HANDLE threadHandle = 0;

struct ThreadParams
{
    int argc;
    LPTSTR* argv;
};


//=========================================================
// JVM-related functions
//

#define NULL_CHECK0(e) if ((e) == 0) return 0
#define NULL_CHECK(e) if ((e) == 0) return

static void AddOption(char *str, void *info);
static void SetClassPath(char *s);
static jboolean ParseArguments(int *pargc, char ***pargv, char **pjarfile,
                               char **pclassname, int *pret);
static jboolean InitializeJVM(JavaVM **pvm, JNIEnv **penv,
                              InvocationFunctions *ifn);
static jstring NewPlatformString(JNIEnv *env, char *s);
static jobjectArray NewPlatformStringArray(JNIEnv *env, char **strv, int strc);
static jstring NewPlatformString(JNIEnv *env, char *s);
static jclass LoadClass(JNIEnv *env, char *name);
static jstring GetMainClassName(JNIEnv *env, char *jarname);

/*
 * Adds a new VM option with the given given name and value.
 */
static void
AddOption(char *str, void *info)
{
    cout << "Adding " << str << endl;

    /*
     * Expand options array if needed to accomodate at least one more
     * VM option.
     */
    if (numOptions >= maxOptions)
    {
        if (options == 0)
        {
            maxOptions = 4;
            options = new JavaVMOption[maxOptions];
        }
        else
        {
            JavaVMOption *tmp;
            maxOptions *= 2;
            tmp = new JavaVMOption[maxOptions];
            memcpy(tmp, options, numOptions * sizeof(JavaVMOption));
            delete [] options;
            options = tmp;
        }
    }
    options[numOptions].optionString = str;
    options[numOptions++].extraInfo = info;
}

static void
SetClassPath(char *s)
{
    char *def = new char[strlen(s) + 40];
    sprintf(def, "-Djava.class.path=%s", s);
    AddOption(def, NULL);
}

/*
 * Returns a new Java string object for the specified platform string.
 */
static jstring
NewPlatformString(JNIEnv *env, char *s)
{
    int len = strlen(s);
    jclass cls;
    jmethodID mid;
    jbyteArray ary;

    NULL_CHECK0(cls = env->FindClass("java/lang/String"));
    NULL_CHECK0(mid = env->GetMethodID(cls, "<init>", "([B)V"));
    ary = env->NewByteArray(len);
    if (ary != 0)
    {
        jstring str = 0;
        env->SetByteArrayRegion(ary, 0, len, (jbyte *)s);
        if (!env->ExceptionOccurred())
        {
            str = (jstring)env->NewObject(cls, mid, ary);
        }
        env->DeleteLocalRef(ary);
        return str;
    }
    return 0;
}

/*
 * Parses command line arguments.
 */
static jboolean
ParseArguments(int *pargc, char ***pargv, char **pjarfile,
               char **pclassname, int *pret)
{
    int argc = *pargc;
    char **argv = *pargv;
    jboolean jarflag = JNI_FALSE;
    char *arg;

    *pret = 1;
    while ((arg = *argv) != 0 && *arg == '-')
    {
        argv++; --argc;
        if (strcmp(arg, "-classpath") == 0 || strcmp(arg, "-cp") == 0)
        {
            if (argc < 1)
            {
                fprintf(stderr, "%s requires class path specification\n", arg);
                return JNI_FALSE;
            }
            SetClassPath(*argv);
            argv++; --argc;
        }
        else if (strcmp(arg, "-jar") == 0)
        {
            jarflag = JNI_TRUE;
        }
        else if (strcmp(arg, "-help") == 0 ||
                 strcmp(arg, "-h") == 0 ||
                 strcmp(arg, "-?") == 0)
        {
            *pret = 0;
            return JNI_FALSE;
        }
        else if (strcmp(arg, "-version") == 0)
        {
            return JNI_FALSE;
        }
        else if (strcmp(arg, "-X") == 0)
        {
            return JNI_FALSE;
        }
        else if (strcmp(arg, "-verbosegc") == 0)
        {
            AddOption("-verbose:gc", NULL);
        }
        else if (strcmp(arg, "-t") == 0)
        {
            AddOption("-Xt", NULL);
        }
        else if (strcmp(arg, "-tm") == 0)
        {
            AddOption("-Xtm", NULL);
        }
        else if (strcmp(arg, "-debug") == 0)
        {
            AddOption("-Xdebug", NULL);
        }
        else if (strcmp(arg, "-noclassgc") == 0)
        {
            AddOption("-Xnoclassgc", NULL);
        }
        else if (strcmp(arg, "-Xfuture") == 0)
        {
            AddOption("-Xverify:all", NULL);
        }
        else if (strcmp(arg, "-verify") == 0)
        {
            AddOption("-Xverify:all", NULL);
        }
        else if (strcmp(arg, "-verifyremote") == 0)
        {
            AddOption("-Xverify:remote", NULL);
        }
        else if (strcmp(arg, "-noverify") == 0)
        {
            AddOption("-Xverify:none", NULL);
        }
        else if (strncmp(arg, "-prof", 5) == 0)
        {
            char *p = arg + 5;
            char *tmp = new char[strlen(arg) + 50];
            if (*p)
            {
                sprintf(tmp, "-Xrunhprof:cpu=old,file=%s", p + 1);
            }
            else
            {
                sprintf(tmp, "-Xrunhprof:cpu=old,file=java.prof");
            }
            AddOption(tmp, NULL);
        }
        else if (strncmp(arg, "-ss", 3) == 0 ||
                 strncmp(arg, "-oss", 4) == 0 ||
                 strncmp(arg, "-ms", 3) == 0 ||
                 strncmp(arg, "-mx", 3) == 0)
        {
            char *tmp = new char[strlen(arg) + 6];
            sprintf(tmp, "-X%s", arg + 1); /* skip '-' */
            AddOption(tmp, NULL);
        }
        else if (strcmp(arg, "-checksource") == 0 ||
                 strcmp(arg, "-cs") == 0 ||
                 strcmp(arg, "-noasyncgc") == 0)
        {
            /* No longer supported */
            fprintf(stderr,
                    "Warning: %s option is no longer supported.\n",
                    arg);
        }
        else
        {
            AddOption(arg, NULL);
        }
    }

    if (--argc >= 0)
    {
        if (jarflag)
        {
            *pjarfile = *argv++;
            *pclassname = 0;
        }
        else
        {
            *pjarfile = 0;
            *pclassname = *argv++;
        }
        *pargc = argc;
        *pargv = argv;
    }

    return JNI_TRUE;
}

/*
 * Initializes the Java Virtual Machine. Also frees options array when
 * finished.
 */
static jboolean
InitializeJVM(JavaVM **pvm, JNIEnv **penv, InvocationFunctions *ifn)
{
    JavaVMInitArgs args;
    jint r;

    memset(&args, 0, sizeof(args));
    args.version  = JNI_VERSION_1_2;
    args.nOptions = numOptions;
    args.options  = options;
    args.ignoreUnrecognized = JNI_FALSE;

    if (debug)
    {
        int i = 0;
        printf("JavaVM args:\n    ");
        printf("version 0x%08lx, ", args.version);
        printf("ignoreUnrecognized is %s, ",
               args.ignoreUnrecognized ? "JNI_TRUE" : "JNI_FALSE");
        printf("nOptions is %ld\n", args.nOptions);
        for (i = 0; i < numOptions; i++)
            printf("    option[%2d] = '%s'\n",
                   i, args.options[i].optionString);
    }

    r = ifn->CreateJavaVM(pvm, (void **)penv, &args);
    free(options);
    return r == JNI_OK;
}


#define NULL_CHECK0(e) if ((e) == 0) return 0
#define NULL_CHECK(e) if ((e) == 0) return

/*
 * Loads a class, convert the '.' to '/'.
 */
static jclass
LoadClass(JNIEnv *env, char *name)
{
    char *buf = new char[strlen(name) + 1];
    char *s = buf, *t = name, c;
    jclass cls;
    jlong start, end;

    cout << "Loading " << name << endl;
    if (debug)
        start = CounterGet();

    do
    {
        c = *t++;
        *s++ = (c == '.') ? '/' : c;
    } while (c != '\0');
    cls = env->FindClass(buf);
    cout << "cls = " << cls << endl;
    delete [] buf;

    if (debug)
    {
        end   = CounterGet();
        printf("%ld micro seconds to load main class\n",
               (jint)Counter2Micros(end-start));
        printf("----_JAVA_LAUNCHER_DEBUG----\n");
    }

    return cls;
}

/*
 * Returns the main class name for the specified jar file.
 */
static jstring
GetMainClassName(JNIEnv *env, char *jarname)
{
#define MAIN_CLASS "Main-Class"
    jclass cls;
    jmethodID mid;
    jobject jar, man, attr;
    jstring str, result = 0;

    NULL_CHECK0(cls = env->FindClass("java/util/jar/JarFile"));
    NULL_CHECK0(mid = env->GetMethodID(cls, "<init>",
                                       "(Ljava/lang/String;)V"));
    NULL_CHECK0(str = NewPlatformString(env, jarname));
    NULL_CHECK0(jar = env->NewObject(cls, mid, str));
    NULL_CHECK0(mid = env->GetMethodID(cls, "getManifest",
                                       "()Ljava/util/jar/Manifest;"));
    man = env->CallObjectMethod(jar, mid);
    if (man != 0)
    {
        NULL_CHECK0(mid = env->GetMethodID(env->GetObjectClass(man),
                                           "getMainAttributes",
                                           "()Ljava/util/jar/Attributes;"));
        attr = env->CallObjectMethod(man, mid);
        if (attr != 0)
        {
            NULL_CHECK0(mid = env->GetMethodID(env->GetObjectClass(attr),
                                               "getValue",
                                               "(Ljava/lang/String;)Ljava/lang/String;"));
            NULL_CHECK0(str = NewPlatformString(env, MAIN_CLASS));
            result = (jstring)env->CallObjectMethod(attr, mid, str);
        }
    }
    return result;
}

/*
 *
 */
bool ExtractMethodIDs()
{
    startMethod = 
        env->GetStaticMethodID(serviceClass, "start", 
                               "([Ljava/lang/String;)V");
    stopMethod =
        env->GetStaticMethodID(serviceClass, "stop", "(V)V");
    pauseMethod =
        env->GetStaticMethodID(serviceClass, "pause", "(V)V");
    resumeMethod =
        env->GetStaticMethodID(serviceClass, "resume", "(V)V");

    if (!startMethod || !stopMethod)
    {
        return false;
    }

    if (!pauseMethod || resumeMethod)
    {
        pauseMethod = resumeMethod = (jmethodID)0;
    }

    return true;
}
/*
 *
 */
boolean BuildJVM(int argc, char* argv[], vector<string>& classArgs)
{
    char *jarfile = 0;
    char *classname = 0;
    char *s = 0;
    char *jvmtype = 0;
    bool jvmspecified = JNI_FALSE;     // Assume no option specified
    int ret;
    InvocationFunctions ifn;
    jlong start, end;

    // Set up debug flag
    if (getenv("_JAVA_LAUNCHER_DEBUG") != 0)
    {
        debug = true;
        printf("----_JAVA_LAUNCHER_DEBUG----\n");
    }

    // Did the user pass a -classic or -hotspot as the first option to
    // the launcher?
    if (argc > 1)
    {
        if (strcmp(argv[1], "-hotspot") == 0)
        {
            jvmtype = "hotspot";
            jvmspecified = true;
        }
        else if (strcmp(argv[1], "-classic") == 0)
        {
            jvmtype = "classic";
            jvmspecified = true;
        }
    }
    ifn.CreateJavaVM = 0;
    ifn.GetDefaultJavaVMInitArgs = 0;
    
    if (!LoadJavaVM(jvmtype, &ifn))
        return false;

    // Grab the program name
    char* progname = *argv++;
    if ((s = strrchr(progname, FILE_SEPARATOR)) != 0)
    {
        progname = s + 1;
    }
    --argc;

    // Skip over a specified -classic/-hotspot option
    if (jvmspecified)
    {
        argv++;
        argc--;
    }

    // Set default CLASSPATH
    if ((s = getenv("CLASSPATH")) == 0)
    {
        s = ".";
    }
    SetClassPath(s);

    // Parse command line options
    if (!ParseArguments(&argc, &argv, &jarfile, &classname, &ret))
    {
        return false;
    }

    // Override class path if -jar flag was specified
    if (jarfile != 0)
    {
        SetClassPath(jarfile);
    }

    // Initialize the virtual machine
    if (debug)
        start = CounterGet();
    if (!InitializeJVM(&vm, &env, &ifn))
    {
        fprintf(stderr, "Could not create the Java virtual machine.\n");
        return false;
    }

    if (debug)
    {
        end   = CounterGet();
        printf("%ld micro seconds to InitializeJVM\n",
               (jint)Counter2Micros(end-start));
    }

    // At this stage, argc/argv have the applications' arguments
    if (debug)
    {
        int i = 0;
        printf("Main-Class is '%s'\n", classname ? classname : "");
        printf("Apps' argc is %d\n", argc);
        for (; i < argc; i++)
        {
            printf("    argv[%2d] = '%s'\n", i, argv[i]);
        }
    }

    // Get the application's main class
    if (jarfile != 0)
    {
        jstring mainClassName = GetMainClassName(env, jarfile);
        if (mainClassName == NULL)
        {
            fprintf(stderr, "Failed to load Main-Class manifest attribute "
                    "from\n%s\n", jarfile);
            vm->DestroyJavaVM();
            return false;
        }
        if (env->ExceptionOccurred())
        {
            env->ExceptionDescribe();
            vm->DestroyJavaVM();
            return false;
        }
        classname = (char *)env->GetStringUTFChars(mainClassName, 0);
        if (classname == NULL)
        {
            env->ExceptionDescribe();
            vm->DestroyJavaVM();
            return false;
        }
        cout << "Loading " << classname << endl;
        serviceClass = LoadClass(env, classname);
        env->ReleaseStringUTFChars(mainClassName, classname);
    }
    else
    {
        serviceClass = LoadClass(env, classname);
    }
    if (serviceClass == NULL)
    {
        env->ExceptionDescribe();
        vm->DestroyJavaVM();
        return false;
    }

    // Extract remainder of args to passed-in vector<string>
    cout << "BuildJVM Getting remainder of args" << endl;    
    for (int i = 0; i < argc; i++)
    {
        string str(*argv++);
        classArgs.push_back(str);
        cout << str << endl;
    }

    return true;
}





//=========================================================
// NTService-related functions
//

/*
 *
 */
DWORD ServiceThread(LPVOID param)
{
    ThreadParams* args = static_cast<ThreadParams*>(param);

    // Construct the JVM & load Java application class
    vector<string> classArgs;
    if (!BuildJVM(args->argc, args->argv, classArgs))
    {
        cout << "BuildJVM failed" << endl;
        return -1;
    }

    // Get our JNI methodIDs
    ExtractMethodIDs();

    // Build args to start(): "java.lang.String[]"
    jvalue call_args[1];

    jclass cls_java_lang_String = LoadClass(env, "java/lang/String");
    jobject defaultValue = env->NewStringUTF("");
    jobjectArray java_args = 
        env->NewObjectArray(classArgs.size(), 
                            cls_java_lang_String, defaultValue);

    // Populate "java_args"
    int i=0;
    for (vector<string>::iterator iter = classArgs.begin();
        iter != classArgs.end();
        iter++)
    {
        jstring str = env->NewStringUTF((*iter).c_str());
        env->SetObjectArrayElement(java_args, i++, str);
    }

    call_args[0].l = java_args;

    // Call "start"
    env->CallStaticVoidMethodA(serviceClass, startMethod, call_args);

    // Clean up
    delete args;

    // Return
    return 0;
}
/*
 *
 */
BOOL InitService(int argc, LPTSTR* argv)
{
    DWORD id;

    ThreadParams* tp = new ThreadParams();
    tp->argc = argc;
    tp->argv = argv;

    threadHandle =
    ::CreateThread(0, 0, (LPTHREAD_START_ROUTINE)ServiceThread,
                   0, 0, &id);
    if (threadHandle == 0)
    {
        return FALSE;
    }
    else
    {
        running = TRUE;
        return TRUE;
    }
}
/*
 *
 */
VOID StopService()
{
    running = FALSE;

    // Report any uncaught Java exceptions thrown from within the JVM
    jthrowable ex;
    if ((ex = env->ExceptionOccurred()) != NULL)
    {
        env->ExceptionDescribe();
    }

    // Destroy the JVM
    vm->DestroyJavaVM();

    ::SetEvent(terminateEvent);
}
/*
 *
 */
VOID PauseService()
{
    paused = TRUE;

    // Call "pause"
    env->CallStaticVoidMethodA(serviceClass,
                               pauseMethod,
                               (jvalue*)0);
}
/*
 *
 */
VOID ResumeService()
{
    paused = FALSE;

    // Call "resume"
    env->CallStaticVoidMethodA(serviceClass,
                               resumeMethod,
                               (jvalue*)0);
}


BOOL SendStatus(DWORD dwCurrState, DWORD dwWin32ExitCode,
                DWORD dwServiceSpecificExitCode, DWORD dwCheckPt,
                DWORD dwWaitHint)
{
    SERVICE_STATUS svcStatus;
    svcStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
    svcStatus.dwCurrentState = dwCurrState;

    if (dwCurrState == SERVICE_START_PENDING)
    {
        svcStatus.dwControlsAccepted = 0;
    }
    else
    {
        svcStatus.dwControlsAccepted =
        SERVICE_ACCEPT_STOP |
        SERVICE_ACCEPT_PAUSE_CONTINUE |
        SERVICE_ACCEPT_SHUTDOWN;
    }

    if (dwServiceSpecificExitCode == 0)
    {
        svcStatus.dwWin32ExitCode = dwWin32ExitCode;
    }
    else
    {
        svcStatus.dwWin32ExitCode = ERROR_SERVICE_SPECIFIC_ERROR;
    }
    svcStatus.dwServiceSpecificExitCode =
    dwServiceSpecificExitCode;

    svcStatus.dwCheckPoint = dwCheckPt;
    svcStatus.dwWaitHint = dwWaitHint;

    BOOL success = ::SetServiceStatus(serviceStatusHandle,
                                      &svcStatus);
    if (!success)
    {
        StopService();
    }

    return success;
}

VOID __stdcall ServiceCtrlHandler(DWORD ctrlCode)
{
    DWORD currState = 0;

    switch (ctrlCode)
    {
        // Do nothing on START
        
        case SERVICE_CONTROL_STOP:
        {
            currState = SERVICE_STOP_PENDING;
            BOOL success = SendStatus(SERVICE_STOP_PENDING, NO_ERROR,
                                      0, 1, 5000);

            StopService();
            break;
        }

        case SERVICE_CONTROL_PAUSE:
        {
            if (running && !paused)
            {
                BOOL success = SendStatus(SERVICE_PAUSE_PENDING, 
                                          NO_ERROR, 0, 1, 5000);

                PauseService();

                currState = SERVICE_PAUSED;

                SendStatus(currState, NO_ERROR, 0, 0, 0);
            }
            break;
        }

        case SERVICE_CONTROL_CONTINUE:
        {
            if (running && paused)
            {
                BOOL success = SendStatus(SERVICE_CONTINUE_PENDING,
                                          NO_ERROR, 0, 1, 1000);
                ResumeService();

                currState = SERVICE_RUNNING;

                SendStatus(currState, NO_ERROR, 0, 0, 0);
            }
            break;
        }

        case SERVICE_CONTROL_INTERROGATE:
        {
            SendStatus(currState, NO_ERROR, 0, 0, 0);
            break;
        }

        case SERVICE_CONTROL_SHUTDOWN:
        {
            break;
        }

        default:
            break;
    }
}

VOID terminate(DWORD err)
{
    if (terminateEvent)
    {
        ::CloseHandle(terminateEvent);
    }

    if (serviceStatusHandle)
    {
        SendStatus(SERVICE_STOPPED, err, 0, 0, 0);
    }

    if (threadHandle)
    {
        ::CloseHandle(threadHandle);
    }
}

VOID __stdcall ServiceMain(DWORD argc, LPTSTR* argv)
{
    // Register with Service Manager
    serviceStatusHandle = 
    ::RegisterServiceCtrlHandler(SERVICE_NAME, 
                                 ServiceCtrlHandler);
    if (!serviceStatusHandle)
    {
        terminate(::GetLastError());
        return;
    }

    BOOL success = SendStatus(SERVICE_START_PENDING, NO_ERROR,
                              0, 1, 5000);
    if (!success)
    {
        terminate(::GetLastError());
        return;
    }

    // Set up terminate signal event
    terminateEvent = ::CreateEvent(0, TRUE, FALSE, 0);
    if (!terminateEvent)
    {
        terminate(::GetLastError());
        return;
    }

    success = SendStatus(SERVICE_START_PENDING, NO_ERROR,
                         0, 2, 5000);
    if (!success)
    {
        terminate(::GetLastError());
        return;
    }

    // Start the service
    success = InitService(argc, argv);
    if (!success)
    {
        terminate(::GetLastError());
        return;
    }

    success = SendStatus(SERVICE_RUNNING, NO_ERROR, 0, 0, 0);
    if (!success)
    {
        terminate(::GetLastError());
        return;
    }

    // Block this thread until stop event is signalled
    ::WaitForSingleObject(terminateEvent, INFINITE);

    terminate(0);
}


int install(int argc, char* argv[])
{
    if (argc != 5)
    {
        cout << "Usage:\n"
            "\t" << argv[0] << " install <name> <label> <cmd-line>\n"
            "\n"
            "Where <name> is the name used internally by the SCM, "
            "<label> is the name appearing in the Services "
            "ControlPanel dialog, "
            "and <cmd-line> is the usual 'java' command-line"
            << endl;

        return 0;
    }

    string name(argv[2]);
    string label(argv[3]);
    string cmdLine(argv[4]);

    cout << "Starting..." << endl;
    cout << "Installing: " << name.c_str() << ", " << 
    label.c_str() << ", " << cmdLine.c_str() << endl;

    SC_HANDLE scm = ::OpenSCManager(0, 0, SC_MANAGER_CREATE_SERVICE);
    if (!scm)
    {
        DWORD err = ::GetLastError();
        cout << "In OpenSCManager: " << err << endl;
        ::ExitProcess(err);
    }

    SC_HANDLE newService =
    ::CreateService(scm, name.c_str(), label.c_str(), 
                    SERVICE_ALL_ACCESS, SERVICE_WIN32_OWN_PROCESS,
                    SERVICE_DEMAND_START, SERVICE_ERROR_NORMAL,
                    argv[0],
                    0, 0, 0, 0, 0);
    if (!newService)
    {
        DWORD err = ::GetLastError();
        cout << "In OpenSCManager: " << err << endl;
        ::ExitProcess(err);
    }

    // Need to configure cmd-line

    cout << "Service '" << name.c_str() << "' installed" << endl;

    ::CloseServiceHandle(newService);
    ::CloseServiceHandle(scm);

    return 0;
}

int remove(int argc, char* argv[])
{
    return 0;
}

int test(int argc, char* argv[])
{
    ThreadParams* tp = new ThreadParams;
    tp->argc = argc;
    tp->argv = argv;
    return ServiceThread(tp);
}


int main(int argc, char** argv)
{
    cout << "Entering main()" << endl;

    // Check args
    if (argc > 1)
    {
        if (strcmp(argv[1], "install")==0)
        {
            install(argc, argv);
        }
        else if (strcmp(argv[1], "remove")==0)
        {
            remove(argc, argv);
        }
        else if (strcmp(argv[1], "test")==0)
        {
            cout << "Testing..." << endl;

            char* argv[] =
            {
                "JVMService.exe",
                "-Dverbose:jni",
                "-Dprop=value",
                "Hello",
                "JoeBob"
            };

            test(5, argv);
            //test(argc, argv);
        }
    }
    else
    {
        // We're being started from the Service Control Manager ?
        // 

        SERVICE_TABLE_ENTRY serviceTable[] =
        {
            { SERVICE_NAME, (LPSERVICE_MAIN_FUNCTION)ServiceMain},
            { 0, 0}
        };

        BOOL success = ::StartServiceCtrlDispatcher(serviceTable);
        if (!success)
        {
            DWORD err = ::GetLastError();

            cout << "In StartServiceCtrlDispatcher" << endl;
            cout << "Err #: " << err << endl;
            ::ExitProcess(err);
        }
    }

    return 0;
}