import java.io.*;

/**
 * This class serves the same purpose as the Java Socket class;
 * a simple abstraction of connecting to and using a named pipe.
 */
public class NTNamedPipe
{
    // Constants
    //
    public static final int PIPE_ACCESS_DUPLEX = 3;
    public static final int PIPE_ACCESS_OUTBOUND = 2;
    public static final int PIPE_ACCESS_INBOUND = 1;
    public static final int PIPE_TYPE_MESSAGE = 4;
    public static final int PIPE_TYPE_BYTE = 0;
    public static final int PIPE_READMODE_MESSAGE = 2;
    public static final int PIPE_READMODE_BYTE = 0;
    public static final int PIPE_WAIT = 0;


    // Internal members
    //
    private int m_hPipe;
    private int m_openMode;


    /**
     *
     */
    public NTNamedPipe()
    { }
    /**
     *
     */
    public NTNamedPipe(String pipeName, int openMode, int pipeMode,
                       int maxInstances, int outBufferSize,
                       int inBufferSize, int timeout)
        throws IOException
    {
        m_openMode = openMode;

        open(pipeName, openMode, pipeMode, maxInstances, 
             outBufferSize, inBufferSize, timeout);

        connect();
    }


    /**
     *
     */
    public void open(String pipeName, int openMode, 
                        int pipeMode, int maxInstances, 
                        int outBufferSize, int inBufferSize, 
                        int timeout)
        throws IOException
    {
        m_hPipe = nativeConstruct(pipeName, openMode, pipeMode,
                                  maxInstances, outBufferSize,
                                  inBufferSize, timeout);
        if (m_hPipe < 0)
        {
            throw new IOException("NT NamedPipe error");
        }
    }
    /**
     *
     */
    public boolean connect()
    {
        return nativeConnect(m_hPipe);
    }
    /**
     *
     */
    public void close()
    {
        nativeClose(m_hPipe);
    }


    /**
     *
     */
    public OutputStream getOutputStream()
        throws IOException
    {
        if ( (m_openMode & PIPE_ACCESS_OUTBOUND) > 0 ||
             (m_openMode & PIPE_ACCESS_DUPLEX) > 0)
        {
            return new NamedPipeOutputStream(m_hPipe);
        }
        else
            throw new IOException("Named-Pipe is inbound only");
    }
    /**
     *
     */
    public InputStream getInputStream()
        throws IOException
    {
        if ( (m_openMode & PIPE_ACCESS_INBOUND) > 0 ||
             (m_openMode & PIPE_ACCESS_DUPLEX) > 0)
        {
            return new NamedPipeInputStream(m_hPipe);
        }
        else
            throw new IOException("Named-Pipe is outbound only");
    }


    // Native methods
    //
    static private native
        int nativeConstruct(String pipeName, int openMode,
                            int pipeMode, int maxInst,
                            int outBufSz, int inBufSz, int to);
    static private native    
        boolean nativeConnect(int pipeHandle);
    
    static private native    
        boolean nativeClose(int pipeHandle);
}


/**
 *
 */
class NamedPipeOutputStream extends OutputStream
{
    // Internal members
    //
    private int m_hPipe;


    /**
     *
     */
    public NamedPipeOutputStream(int hPipe)
    {
        m_hPipe = hPipe;
    }


    /**
     *
     */
    public void close()
    {
        // Do nothing--the named pipe may still be open in incoming
        // mode, so we don't want to close it
    }
    /**
     *
     */
    public void flush()
    {
        nativeFlush(m_hPipe);
    }
    /**
     *
     */
    public void write(byte[] b)
    {
        nativeWrite(m_hPipe, b);
    }
    /**
     *
     */
    public void write(int ch)
    {
        nativeWrite(m_hPipe, ch);
    }


    // Native methods
    //
    private static native void nativeFlush(int hPipe);
    private static native void nativeWrite(int hPipe, byte[] bytes);
    private static native void nativeWrite(int hPipe, int ch);
}


/**
 *
 */
class NamedPipeInputStream extends InputStream
{
    // Internal members
    //
    private int m_hPipe;


    /**
     *
     */
    public NamedPipeInputStream(int hPipe)
    {
        m_hPipe = hPipe;
    }


    /**
     *
     */
    public int available()
    {
        return 0;
    }
    /**
     *
     */
    public void close()
    {
        // Do nothing--the named pipe may still be open in outbound
        // mode, so we don't want to close it
    }
    /**
     *
     */
    public boolean markSupported()
    {
        return false;
    }
    /**
     *
     */
    public int read(byte[] b)
    {
        return nativeRead(m_hPipe, b);
    }
    /**
     *
     */
    public int read()
    {
        return nativeRead(m_hPipe);
    }


    // Native methods
    //
    private static native void nativeAvailable(int hPipe);
    private static native int nativeRead(int hPipe, byte[] b);
    private static native int nativeRead(int hPipe);
}
