package com.javageeks.client;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.InterruptedIOException;
import java.io.IOException;
import java.net.Socket;
import java.util.Enumeration;
import java.util.Vector;

/**
 *
 */
public class HeartbeatListenerClient
    implements Runnable
{
    // Internal members
    //
    private String m_host;
    private int m_port;
    private String m_pingMsg;
    private int m_interval;
    private int m_giveUpCount;
    private Thread m_thread = null;
    private Vector m_listeners = new Vector();
    
    
    /**
     * "Complete" constructor--initialize with all given values
     */
    public HeartbeatListenerClient(String host, int port,
                                   String msg, int interval,
                                   int giveUpCount, Thread thread)
    {
        m_host = host;
        m_port = port;
        m_pingMsg = msg;
        m_interval = interval;
        m_giveUpCount = giveUpCount;
        m_thread = thread;
    }
    /**
     * Convenience constructor--assumes defaults of "PING", 15
     * seconds, and 3 attempts
     */
    public HeartbeatListenerClient(String host, int port)
    {
        this(host, port, "PING", 15 * 1000, 3, null);
    }

    
    /**
     * Add a HeartbeatListener to the list of interested parties
     */
    public void addListener(HeartbeatListener listener)
    {
        m_listeners.addElement(listener);
    }
    /**
     * Remove a HeartbeatListener from the list of notification
     * targets on heartbeat pings or failures
     */
    public void removeListener(HeartbeatListener listener)
    {
        m_listeners.remove(listener);
    }


    /**
     * Begin listening for heartbeat messages
     */
    public void startListening()
    {
        if (m_thread == null)
        {
            m_thread = new Thread(this);
        }

        m_thread.start();
    }
    /**
     * Cease listening for heartbeat messages
     */
    public void stopListening()
    {
        m_thread.interrupt();
    }


    /**
     *
     */
    public void run()
    {
        Socket socket = null;
        try
        {
            socket = new Socket(m_host, m_port);
            BufferedReader fromSocket = new BufferedReader(
                new InputStreamReader(socket.getInputStream()));
            
            // We want to block for only 5 seconds waiting for input
            // (this would be programmatically controlled in other
            // more flexible systems)
            socket.setSoTimeout(m_interval);
            
            // Wait for up to three missed pings before giving up
            int giveUpCount = 0;
            while (giveUpCount < m_giveUpCount)
            {
                try
                {
                    String line = fromSocket.readLine();
                    if (line == null)
                        giveUpCount++;
                    else if (line.equals(m_pingMsg))
                    {
                        // Reset giveUpCount; we got a message
                        giveUpCount = 0;
                        
                        // Broadcast the message on to our listeners
                        for (Enumeration enum = 
                             m_listeners.elements();
                             enum.hasMoreElements(); )
                        {
                            HeartbeatListener l =
                                (HeartbeatListener)enum.nextElement();
                                
                            l.onHeartbeatPing(line);
                        }
                    }
                }
                catch (java.io.InterruptedIOException ex)
                {
                    giveUpCount++;
                }
            }
            
            // If we got here, it's because we gave up
            for (Enumeration enum = m_listeners.elements();
                 enum.hasMoreElements(); )
            {
                HeartbeatListener l =
                    (HeartbeatListener)enum.nextElement();
                    
                l.onHeartbeatFail(m_pingMsg);
            }
        }
        catch (Exception ex)
        {
            // Not a very reusable option, but the only way to know
            // when an Exception is thrown, since we can't throw it
            // out of run (Runnable.run has no "throws" clause)
            ex.printStackTrace();
        }
    }
    
    
    /**
     * Test driver; for testing purposes only.
     */
    public static void main(String[] args)
    {
        if (args.length < 1)
        {
            System.out.println("Usage: java HeartbeatListenerClient"
                               + " <hostname:port>");
            return;
        }
        
        // Parse out hostname and port
        String host = args[0].substring(0, args[0].indexOf(":"));
        Integer port =
            new Integer(args[0].substring(args[0].indexOf(":")+1,
                                             args[0].length()));
        
        System.out.println("Connecting to " + host + ":" + port);

        HeartbeatListenerClient client =
            new HeartbeatListenerClient(host, port.intValue(),
                                        "PING", 15 * 1000, 3, null);
        
        client.addListener(new HeartbeatListener()
        {
            public void onHeartbeatPing(String pingMsg)
            {
                System.out.println("PING!: " + pingMsg);
            }
            public void onHeartbeatFail(String pingMsg)
            {
                System.out.println("Heartbeat's stopped!");
            }
        });

        client.startListening();
    }
}