package com.javageeks.gjas;

import java.io.*;
import java.rmi.*;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.Vector;
import com.javageeks.classloader.ClassLoaderStrategy;
import com.javageeks.classloader.HashtableClassLoader;

/**
 * RMIServerManager is actually a thin Proxy to an RMIServerManagerServer,
 * which exists within some other JVM (which may or may not be on another
 * machine). 
 *
 * <P>Note that throughout this class, rather than throw declared exceptions,
 * we throw RuntimeExceptions (which aren't required to be declared in the
 * methods' "throws" clause). Doing this allows us to maintain the
 * IServerManager interface, and, quite frankly, I never saw RemoteException
 * as a need-to-be-checked Exception type, anyway.
 */
public class RMIServerManager
    implements IServerManager
{
    public RMIServerManager(String host)
        throws Exception
    {
        // Set up ServerManager Singleton
        ServerManager.instance(this);

        // Connect to server; throw RuntimeException if that fails
        //System.out.println("In RMIServerManager--attempting lookup");        
        IRemoteServerManager remoteSvrMgr =
            (IRemoteServerManager)Naming.lookup("rmi://" + host + "/" +
                                                IRemoteServerManager.RMI_LOOKUPNAME);
        //System.out.println("In RMIServerManager--lookup complete");        
        m_rmiSvrMgr = remoteSvrMgr;
    }

    /**
     *
     */
    public void shutdown()
    {
        try
        {
            m_rmiSvrMgr.remoteShutdown();
        }
        catch (java.rmi.RemoteException remoteEx)
        {
            throw new RuntimeException(remoteEx.toString());
        }
    }
    

    /**
     *
     */
    public void deployService(String serviceName, 
                              ClassLoaderStrategy strategy)
    {
        try
        {
            m_rmiSvrMgr.remoteDeployService(serviceName, strategy);
        }
        catch (java.rmi.RemoteException remoteEx)
        {
            throw new RuntimeException(remoteEx.toString());
        }
    }
    /**
     *
     */
    public IServer loadService(String svcName)
    {
        try
        {
            return m_rmiSvrMgr.remoteLoadService(svcName);
        }
        catch (java.rmi.RemoteException remoteEx)
        {
            throw new RuntimeException(remoteEx.toString());
        }
    }
    /**
     *
     */
    public IServer loadService(Service svc)
    {
        try
        {
            return m_rmiSvrMgr.remoteLoadService(svc);
        }
        catch (java.rmi.RemoteException remoteEx)
        {
            throw new RuntimeException(remoteEx.toString());
        }
    }
    /**
     *
     */
    public IServer addService(String svcName, 
                              ConfigProperties args)
    {
        try
        {
            return m_rmiSvrMgr.remoteAddService(svcName, args);
        }
        catch (java.rmi.RemoteException remoteEx)
        {
            throw new RuntimeException(remoteEx.toString());
        }
    }
    /**
     *
     */
    public IServer addService(Service svc, 
                              ConfigProperties args)
    {
        try
        {
            return m_rmiSvrMgr.remoteAddService(svc, args);
        }
        catch (java.rmi.RemoteException remoteEx)
        {
            throw new RuntimeException(remoteEx.toString());
        }
    }
    /**
     *
     */
    public void removeService(String instanceID)
    {
        try
        {
            m_rmiSvrMgr.remoteRemoveService(instanceID);
        }
        catch (java.rmi.RemoteException remoteEx)
        {
            throw new RuntimeException(remoteEx.toString());
        }
    }
    /**
     *
     */
    public void killService(String instanceID)
    {
        try
        {
            m_rmiSvrMgr.remoteKillService(instanceID);
        }
        catch (java.rmi.RemoteException remoteEx)
        {
            throw new RuntimeException(remoteEx.toString());
        }
    }
    /**
     *
     */
    public String[] getServices()
    {
        try
        {
            return m_rmiSvrMgr.remoteGetServices();
        }
        catch (java.rmi.RemoteException remoteEx)
        {
            throw new RuntimeException(remoteEx.toString());
        }
    }
    /**
     *
     */
    public IServer getService(String instanceID)
    {
        try
        {
            return m_rmiSvrMgr.remoteGetService(instanceID);
        }
        catch (java.rmi.RemoteException remoteEx)
        {
            throw new RuntimeException(remoteEx.toString());
        }
    }
    
    
   /**
     *
     */
    public void log(String msg)
    {
        try
        {
            m_rmiSvrMgr.remoteLog(msg);
        }
        catch (java.rmi.RemoteException remoteEx)
        {
            throw new RuntimeException(remoteEx.toString());
        }
    }
    /**
     *
     */
    public void log(Exception ex)
    {
        try
        {
            m_rmiSvrMgr.remoteLog(ex);
        }
        catch (java.rmi.RemoteException remoteEx)
        {
            throw new RuntimeException(remoteEx.toString());
        }
    }
    /**
     *
     */
    public void error(String msg)
    {
        try
        {
            m_rmiSvrMgr.remoteError(msg);
        }
        catch (java.rmi.RemoteException remoteEx)
        {
            throw new RuntimeException(remoteEx.toString());
        }
    }
    /**
     *
     */
    public void error(Exception ex)
    {
        try
        {
            m_rmiSvrMgr.remoteError(ex);
        }
        catch (java.rmi.RemoteException remoteEx)
        {
            throw new RuntimeException(remoteEx.toString());
        }
    }


    public static void main (String args[])
        throws Exception
    {
        if (args.length < 1)
        {
            System.out.println("Usage: java RMIServerManager <hostname>");
            return;
        }

        // Create (and register) the RMIServerManager
        new RMIServerManager(args[0]);

        ServerManager.log("Entering RMIServerManager.main()");

        // Parse command-line arguments, if any
        //
        for (int argc=1; argc < args.length; argc++)
        {
            if (args[argc].startsWith("@"))
            {
                // The "@" argument indicates the file we should
                // parse for services to execute
                try
                {
                    String arg = args[argc];
                    String filename = 
                        arg.substring(arg.indexOf("@")+1, arg.length());
                    FileInputStream fis = 
                        new FileInputStream(filename);
                    ServerManager.parseInputStream(fis);
                }
                catch (Exception ex)
                {
                    // Ignore it and move on

                    ex.printStackTrace();
                }
            }
            else if ("TEST".equals(args[argc]))
            {
                // Deploy a Service, then try to add it.
                try
                {
                    // Look for "TestService.class" in the current
                    // directory
                    String filename = "TestService.class";
                    java.io.FileInputStream fis =
                        new java.io.FileInputStream(filename);

                    byte[] bytes = new byte[fis.available()];
                    fis.read(bytes);

                    // Create a HashtableClassLoader
                    com.javageeks.classloader.HashtableClassLoader
                        hcl = new HashtableClassLoader();
                    hcl.putClass("TestService", bytes);

                    // Deploy it
                    ServerManager.deployService("TestService", hcl);

                    // Now add the Service
                    IServer svr = 
                        ServerManager.loadService("TestService");
                    if (svr == null)
                    {
                        System.out.println("Test failed!");
                    }

                    System.out.println("Calling start....");
                    ConfigProperties props = svr.getConfigInfo();
                    for (Iterator i = props.properties(); i.hasNext(); )
                    {
                        System.out.println(((ConfigProperty)i.next()).toString());
                    }

                    svr.start();
                }
                catch (Exception ex)
                {
                    ex.printStackTrace();
                }
            }
            else
            {
                ServerManager.parseArg(args[argc]);
            }
        }
            
        ServerManager.log("Exiting RMIServerManager.main()");
    }


    // Internal data
    //
    IRemoteServerManager m_rmiSvrMgr;
}
