package com.javageeks.gjas;

import java.rmi.*;
import java.rmi.server.*;
import java.util.HashMap;
import com.javageeks.classloader.ClassLoaderStrategy;

/**
 * 
 */
public class RMIServerManagerServer extends UnicastRemoteObject
    implements IRemoteServerManager
{
    /**
     *
     */
    public RMIServerManagerServer(IServerManager svrMgr)
        throws RemoteException
    {
        m_serverMgr = svrMgr;
    }


    //===========================================
    // IRemoteServerManager-inherited methods (implementations)
    //
    
    /**
     * Shut the entire system down, usually in preparation for terminating
     * this VM (or perhaps for doing a complete shutdown/restart cycling).
     * Effectively, this is the same as calling <code>getServices</code> to get all
     * Servers' instanceIDs, then calling <code>removeService</code> on each one.
     */
    public void remoteShutdown()
        throws RemoteException
    {
        m_serverMgr.shutdown();
    }
    

    /**
     * "Deploy" this Service to the ServerManager
     */
    public void remoteDeployService(String serviceName, 
                                    ClassLoaderStrategy strategy)
        throws RemoteException
    {
        m_serverMgr.deployService(serviceName, strategy);
    }
    /**
     * Add this Service from the deployed Services
     */
    public IServer remoteLoadService(String svcName)
        throws RemoteException
    {
        // Do the normal addService thing
        IServer svr = m_serverMgr.loadService(svcName);

        // Create our RMI Proxy
        long ID = m_serverCt++;
        RMIServer rmiSvr = new RMIServer(this, ID);
        m_serverMap.put(Long.toString(ID), svr);
            // Careful readers will note that this has an inherent design flaw;
            // when we get above the maximum count of a long, we will wrap around,
            // with the possibility that an existing Server could be overwritten!
            // However, my experience has led me to believe that very rarely, if
            // ever, will this actually occur. Consider the mathematics--if a new
            // Server was added every second, it would take 2^64 seconds before
            // overlap occurs, and the human civilization hasn't even been in
            // existence that long! Consider the mathematics: 2^32 seconds
            // (4,234,967,296) is 71582788 minutes, or 1193046 hours or
            // 49,710 days, or 136 years! And Java uses 64-bit long types, which
            // is 136 years, squared, or roughly 18,500 years!
        
        return rmiSvr;
    }
    /**
     * Add the loaded Service to the list of Servers and start it. We throw away
     * the return value from the LocalServerManager call (the LocalServer instance)
     * because we need to construct an RMIServer instance to give back to the
     * RMI caller.
     */
    public IServer remoteLoadService(Service svc)
        throws RemoteException
    {
        // Do the normal addService thing
        IServer svr = m_serverMgr.loadService(svc);

        // Create our RMI Proxy
        long ID = m_serverCt++;
        RMIServer rmiSvr = new RMIServer(this, ID);
        m_serverMap.put(Long.toString(ID), svr);

        return rmiSvr;
    }
    /**
     * Add this Service from the deployed Services
     */
    public IServer remoteAddService(String svcName, 
                                    ConfigProperties args)
        throws RemoteException
    {
        // Do the normal addService thing
        IServer svr = m_serverMgr.addService(svcName, args);

        // Create our RMI Proxy
        long ID = m_serverCt++;
        RMIServer rmiSvr = new RMIServer(this, ID);
        m_serverMap.put(Long.toString(ID), svr);
        
        return rmiSvr;
    }
    /**
     * Add the loaded Service to the list of Servers and start it. We throw away
     * the return value from the LocalServerManager call (the LocalServer instance)
     * because we need to construct an RMIServer instance to give back to the
     * RMI caller.
     */
    public IServer remoteAddService(Service svc, 
                                    ConfigProperties args)
        throws RemoteException
    {
        // Do the normal addService thing
        IServer svr = m_serverMgr.addService(svc, args);

        // Create our RMI Proxy
        long ID = m_serverCt++;
        RMIServer rmiSvr = new RMIServer(this, ID);
        m_serverMap.put(Long.toString(ID), svr);

        return rmiSvr;
    }
    /**
     * Attempt to stop (if necessary) and remove an instance of a Server.
     * Because it's possible that multiple Servers of a given type can be
     * running simultaneously (for example, sockets-based Services listening
     * on multiple ports), we need to have the user identify which Server
     * they wish shut down by using the Server instance's instanceID.
     */
    public void remoteRemoveService(String instanceID)
        throws RemoteException
    {
        // Do the normal removeService thing
        m_serverMgr.removeService(instanceID);
    }
    /**
     * Try to kill the Service--don't try to stop() it
     */
    public void remoteKillService(String instanceID)
        throws RemoteException
    {
        m_serverMgr.killService(instanceID);
    }
    /**
     * Obtain a list of every Server instance running in the system.
     */     
    public String[] remoteGetServices()
        throws RemoteException
    {
        return m_serverMgr.getServices();
    }
    /**
     * Obtain a reference to a Server instance by ID. If it can't be found
     * (perhaps it's shut down since the user obtained the ID?), then return
     * a null instance.
     */
    public IServer remoteGetService(String instanceID)
        throws RemoteException
    {
        IServer svr = m_serverMgr.getService(instanceID);
        
        // Create our RMI Proxy
        long ID = m_serverCt++;
        RMIServer rmiSvr = new RMIServer(this, ID);
        m_serverMap.put(Long.toString(ID), svr);

        return rmiSvr;
    }
    
    
    /**
     *
     */
    public void remoteLog(String msg)
        throws RemoteException
    {
        m_serverMgr.log(msg);
    }
    /**
     *
     */
    public void remoteLog(Exception ex)
        throws RemoteException
    {
        m_serverMgr.log(ex);
    }
    /**
     *
     */
    public void remoteError(String msg)
        throws RemoteException
    {
        m_serverMgr.error(msg);
    }
    /**
     *
     */
    public void remoteError(Exception ex)
        throws RemoteException
    {
        m_serverMgr.error(ex);
    }


    //===========================================
    // IServer-functionality
    //
    /**
     *
     */
    public boolean start(long ID)
        throws RemoteException
    {
        IServer svr = (IServer)m_serverMap.get(Long.toString(ID));
        return svr.start();
    }
    /**
     *
     */
    public boolean stop(long ID)
        throws RemoteException
    {
        IServer svr = (IServer)m_serverMap.get(Long.toString(ID));
        return svr.stop();
    }
    /**
     *
     */
    public boolean pause(long ID)
        throws RemoteException
    {
        IServer svr = (IServer)m_serverMap.get(Long.toString(ID));
        return svr.pause();
    }
    /**
     *
     */
    public boolean resume(long ID)
        throws RemoteException
    {
        IServer svr = (IServer)m_serverMap.get(Long.toString(ID));
        return svr.resume();
    }
    /**
     *
     */
    public void kill(long ID)
        throws RemoteException
    {
        IServer svr = (IServer)m_serverMap.get(Long.toString(ID));
        svr.kill();
    }
        
    /**
     *
     */
    public String getState(long ID)
        throws RemoteException
    {
        IServer svr = (IServer)m_serverMap.get(Long.toString(ID));
        return svr.getState();
    }
        
    /**
     *
     */
    public String getInstanceID(long ID)
        throws RemoteException
    {
        IServer svr = (IServer)m_serverMap.get(Long.toString(ID));
        return svr.getInstanceID();
    }
        
    /**
     *
     */
    public Exception getLastError(long ID)
        throws RemoteException
    {
        IServer svr = (IServer)m_serverMap.get(Long.toString(ID));
        return svr.getLastError();
    }


    /**
     *
     */
    public ConfigProperties getConfigInfo(long ID)
        throws RemoteException
    {
        IServer svr = (IServer)m_serverMap.get(Long.toString(ID));
        return svr.getConfigInfo();
    }
    /**
     *
     */
    public void setConfigInfo(long ID, ConfigProperties info)
        throws RemoteException
    {
        IServer svr = (IServer)m_serverMap.get(Long.toString(ID));
        svr.setConfigInfo(info);
    }



    //===========================================
    // main
    //
    /**
     *
     */
    public static void main (String args[])
        throws Exception
    {
        // Create an instance of RMIServerManagerServer
        RMIServerManagerServer svr = 
            new RMIServerManagerServer(new LocalServerManager());

        // Bind & export it
        Naming.bind(IRemoteServerManager.RMI_LOOKUPNAME, svr);
        svr.remoteLog("RMIServerManagerServer bound to registry");
    }


    // Internal data
    //
    private IServerManager m_serverMgr;
    private long m_serverCt = 0;
    private HashMap m_serverMap = new HashMap();
}
