package com.javageeks.gjas.connection;

import java.io.*;
import java.net.*;
import java.util.*;
import com.javageeks.gjas.ConfigProperties;
import com.javageeks.gjas.ServerManager;

/**
 * A cheap way of doing #defines in Java; just implement this interface
 * (which costs you nothing), and you can refer to them directly
 */
interface HttpConnectionConstants
{
    // 2xx: "OK" response codes
    public static final int HTTP_OK = 200;
    public static final int HTTP_CREATED = 201;
    public static final int HTTP_ACCEPTED = 202;
    public static final int HTTP_NOT_AUTHORITATIVE = 203;
    public static final int HTTP_NO_CONTENT = 204;
    public static final int HTTP_RESET = 205;
    public static final int HTTP_PARTIAL = 206;

    // 3xx: relocation/redirect response codes
    public static final int HTTP_MULT_CHOICE = 300;
    public static final int HTTP_MOVED_PERM = 301;
    public static final int HTTP_MOVED_TEMP = 302;
    public static final int HTTP_SEE_OTHER = 303;
    public static final int HTTP_NOT_MODIFIED = 304;
    public static final int HTTP_USE_PROXY = 305;
    
    // 4xx: client error codes
    public static final int HTTP_BAD_REQUEST = 400;
    public static final int HTTP_UNAUTHORIZED = 401;
    public static final int HTTP_PAYMENT_REQUIRED = 402;
    public static final int HTTP_FORBIDDEN = 403;
    public static final int HTTP_NOT_FOUND = 404;
    public static final int HTTP_BAD_METHOD = 405;
    public static final int HTTP_NOT_ACCEPTABLE = 406;
    public static final int HTTP_PROXY_AUTH = 407;
    public static final int HTTP_CLIENT_TIMEOUT = 408;
    public static final int HTTP_CONFLICT = 409;
    public static final int HTTP_GONE = 410;
    public static final int HTTP_LENGTH_REQUIRED = 411;
    public static final int HTTP_PRECON_FAILED = 412;
    public static final int HTTP_ENTITY_TOO_LARGE = 413;
    public static final int HTTP_REQ_TOO_LONG = 414;
    public static final int HTTP_UNSUPPORTED_TYPE = 415;
    
    // 5xx: server error codes
    public static final int HTTP_SERVER_ERROR = 500;
    public static final int HTTP_INTERNAL_ERROR = 501;
    public static final int HTTP_BAD_GATEWAY = 502;
    public static final int HTTP_UNAVAILABLE = 503;
    public static final int HTTP_GATEWAY_TIMEOUT = 504;
    public static final int HTTP_VERSION = 505;
}    

/**
 * This class provides a simplistic HTTP/1.0 service; it uses only a single
 * root (as opposed to other Web servers, which allow for multiple "virtual
 * roots" in their setup), and offers only the most rudimentary of HTTP
 * protocol services. Because HttpConnection is a transient type (that is,
 * multiple HttpConnection instances will come and go without warning or
 * guarantee), we use static instances to carry expensive resources (like
 * the properties we use) from instance to instance without having to reload
 * them each time.
 *
 * Much of this code is cribbed from the WebServer.java example from the
 * JavaSoft site; however, if you're familiar with that code, you'll notice
 * that this version is much smaller, owing to the fact that GJAS factors out
 * much of the complexity unrelated to HTTP (like thread pools).
 *
 * <P><B>TODO: </B>
 * <UL>
 * <LI>Reread ConfigProperties info in each Connection instance; read
 * .http-properties as "base" defaults overridden by ConfigProperties
 * </UL>
 */
public class HttpConnection
    implements Connection, HttpConnectionConstants
{
    static final String EOL = new String(new byte[] { (byte)'\r', (byte)'\n' });

    /**
     * 
     */
    public void serve(Socket socket, ConfigProperties props)
        throws Exception
    {
        // Before we even do anything, do we need to abort?
        if (s_exception != null)
            throw s_exception;
    
        // Get streams from the Socket
        InputStream in = socket.getInputStream();
        OutputStream out = socket.getOutputStream();
        
        // Pick apart the URL sent; we only want the first line, really
        Properties headers = new Properties();  // HTTP headers we send back

        try
        {
            // Pick out the request, load the content and the headers
            parseRequest(in, headers);
            
            // Send back our HTTP/1.0 response
            sendResponse(out, headers);
        }
        catch (Exception ex)
        {
            // Send the server error back to the client
            sendError(HTTP_SERVER_ERROR, ex.toString(), headers);
            throw ex;
        }
        finally
        {
            // Send it back down the stream
            out.flush();
            in.close();
            out.close();
        }
    }
    
    /**
     *
     */
    public void parseRequest(InputStream in, Properties headers)
        throws IOException
    {
        headers.put("_HTTPVersion", "1.0");
        headers.put("Server", "GJAS-HttpConnection/1.0");
        headers.put("Date", new Date());

        // Pick out the command & URL in the incoming request
        BufferedReader reader = 
            new BufferedReader(new InputStreamReader(in));
        String line;
        String clientCommand = null;
        while ((line = reader.readLine()) != null)
        {
            if (line.startsWith("GET"))
            {
                clientCommand = line;
                break;
            }
            else
            {
                System.out.println(line);
            }
        }
        if (clientCommand == null)
        {
            sendError(HTTP_BAD_METHOD, "Server only supports GET", headers);
            return;
        }
        
        // Pick out the command & URL
        String URL = clientCommand.substring(4);
        URL = URL.substring(0, URL.indexOf(" "));
        
        // Turn the URL into a local filename
        File target;
        if (URL.equals("/"))
        {
            target = new File(s_root.getCanonicalPath());
        }
        else
        {
            String filename = new String(URL.replace('/', File.separatorChar));
            if (filename.startsWith(File.separator))
                filename = filename.substring(1);
            target = new File(s_root, filename);
        }
        
        // Find file/directory
        if (!target.exists())
        {
            sendError(HTTP_BAD_REQUEST, "URL '" + URL + "' not found", headers);
            return;
        }
        if (target.isDirectory())
        {
            File targetIndex = new File(target, "index.html");

            if (targetIndex.exists())
            {
                // Send the file back
                target = targetIndex;

                headers.put("_ReturnCode", HTTP_OK + " OK");
                headers.put("_Content", new FileInputStream(target));
                headers.put("Last-modified", new Date(target.lastModified()));
                headers.put("Content-length", new Long(target.length()));
                headers.put("Content-type", "text/html");
            }
            else
            {
                // Send the contents of the directory back
                String html = listDirectory(URL, target);
                headers.put("_ReturnCode", HTTP_OK + " OK");
                headers.put("_Content",
                    new ByteArrayInputStream(html.getBytes()));
                headers.put("Last-modified", new Date());
                headers.put("Content-length", new Long(html.length()));
                headers.put("Content-type", "text/html");
            }
        }
        else
        {
            // Send the file back
            headers.put("_ReturnCode", HTTP_OK + " OK");
            headers.put("_Content", new FileInputStream(target));
            headers.put("Last-modified", new Date(target.lastModified()));
            headers.put("Content-length", new Long(target.length()));
            String fname = target.getName();
            int ind = fname.lastIndexOf('.');
            if (ind > 0)
                headers.put("Content-type", s_suffixMap.get(fname.substring(ind)));
            else
                headers.put("Content-type", "unknown/unknown");
        }
    }
    private void sendError(int errCode, String errText, Properties headers)
    {
        headers.put("_ReturnCode", errCode + " " + errText);
    }
    private String listDirectory(String URL, File dir)
        throws IOException
    {
        StringBuffer ret = new StringBuffer();
        
        // Header information
        ret.append("<TITLE>Directory of " + URL + "</TITLE>" + "\n<P>\n");
        
        if (!dir.getCanonicalPath().equals(s_root.getCanonicalPath()))
            ret.append("<A HREF=\"..\">Up one directory</A><BR>\n<P>\n\n");
        
        // Print list of files
        String[] list = dir.list();
        for (int i=0; list != null && i<list.length; i++)
        {
            File item = new File(dir, list[i]);
            if (item.isDirectory())
            {
                ret.append("<A HREF=\"" + URL + list[i] + "/\">" + 
                    list[i] + "/</A><BR>\n");
            }
            else
            {
                ret.append("<A HREF=\"" + URL + list[i] + "\">" + 
                    list[i] + "</A><BR>\n");
            }
        }
        
        // Print trailer
        ret.append("<P><HR><BR><I>" + new Date() + "</I><BR>");
        ret.append("<I>Generated by the Generic Java Application Server, " +
            getClass().getName() + " service</I>");
            
        return new String(ret);
    }
    
    /**
     *
     */
    public void sendResponse(OutputStream os, Properties headers)
        throws IOException
    {
        PrintWriter pw = new PrintWriter(os);
    
        // Send HTTP/1.0 <retCode> <text> line
        pw.print("HTTP/" + headers.get("_HTTPVersion") + " " +
            headers.get("_ReturnCode"));
        pw.write(EOL);

        // Send headers, one at a time
        for (Enumeration enum = headers.propertyNames(); 
             enum.hasMoreElements(); )
        {
            String key = (String)enum.nextElement();
            
            // Skip the lines starting with an underscore; those are
            // for HttpConnection internal use only
            if (key.startsWith("_"))
                continue;
            
            // Print the key, a colon, the property, and EOL
            pw.print(key + ": " + headers.get(key) + EOL);
        }

        // Write blank line after headers to mark end of headers
        pw.write(EOL);
        pw.flush();

        // Read the content from the specified InputStream, and
        // send it down
        InputStream content = (InputStream)headers.get("_Content");
        if (content != null)
        {
            byte[] buffer = new byte[1024];
            int nRead = 0;
            while ((nRead = content.read(buffer)) != -1)
            {
                //pw.print(new String(buffer, 0, nRead));
                os.write(buffer, 0, nRead);
            }

            // Write one more blank line, just for luck
            pw.write(EOL);
        }
        
        // Ba-whoosh!
        pw.flush();
    }
    
    
    // Internal members
    //    
    private static Exception s_exception = null;
        // Reserved solely for exceptions generated from the static
        // init block
    private static Properties s_properties = null;
    private static File s_root = null;
    private static Hashtable s_suffixMap = new Hashtable();

    
    // Internal methods
    //
    /**
     * Static initializer block--executed when this class is loaded
     * into its namespace.
     */
    static
    {
        // Load the s_properties instance with our http-properties file
        loadProperties();
        // Fill the suffix map with well-known file types
        fillMap();
    }
            
    /**
     *
     */
    private static void loadProperties()
    {
        try
        {
            // Load the properties from disk
            s_properties = new Properties();
            InputStream propFileStream;
            try
            {
                propFileStream = 
                    new FileInputStream(".http-properties");
                s_properties.load(propFileStream);
                s_properties.list(System.out);
                propFileStream.close();
            }
            catch (IOException ioEx)
            { }

            // Set up our instance data from the Properties read in
            String r = s_properties.getProperty("http-root");
            if (r != null)
            {
                s_root = new File(r);
                if (!s_root.exists())
                {
                    //s_exception = 
                    //    new FileNotFoundException("Root doesn't exist!");
                    s_root = null;
                }
            }
            
            // Use defaults
            if (s_root == null)
                s_root = new File(System.getProperty("user.dir"));
            
            ServerManager.log(
                HttpConnection.class.getName() + ": http-root = " +
                s_root.getCanonicalPath());
        }
        catch (IOException ioEx)
        {
            // Nothing to do, so just move on
            s_exception = ioEx;
            ServerManager.instance().log(ioEx);
        }
    }
    /**
     *
     */
    private static void fillMap()
    {
        s_suffixMap.put("", "content/unknown");
        s_suffixMap.put(".uu", "application/octet-stream");
        s_suffixMap.put(".exe", "application/octet-stream");
        s_suffixMap.put(".ps", "application/postscript");
        s_suffixMap.put(".zip", "application/zip");
        s_suffixMap.put(".sh", "application/x-shar");
        s_suffixMap.put(".tar", "application/x-tar");
        s_suffixMap.put(".snd", "audio/basic");
        s_suffixMap.put(".au", "audio/basic");
        s_suffixMap.put(".wav", "audio/x-wav");
        s_suffixMap.put(".gif", "image/gif");
        s_suffixMap.put(".jpg", "image/jpeg");
        s_suffixMap.put(".jpeg", "image/jpeg");
        s_suffixMap.put(".htm", "text/html");
        s_suffixMap.put(".html", "text/html");
        s_suffixMap.put(".txt", "text/plain");
        s_suffixMap.put(".text", "text/plain");
        s_suffixMap.put(".c", "text/plain");
        s_suffixMap.put(".cc", "text/plain");
        s_suffixMap.put(".cpp", "text/plain");
        s_suffixMap.put(".c++", "text/plain");
        s_suffixMap.put(".h", "text/plain");
        s_suffixMap.put(".hh", "text/plain");
        s_suffixMap.put(".hpp", "text/plain");
        s_suffixMap.put(".h++", "text/plain");
        s_suffixMap.put(".pl", "text/plain");
        s_suffixMap.put(".java", "text/plain");
    }
}    