package com.javageeks.gjas.services;

import com.javageeks.gjas.*;
import com.javageeks.classloader.ClassLoaderStrategy;
import com.javageeks.classloader.StrategyClassLoader;
import com.javageeks.gjas.connection.Connection;
import EDU.oswego.cs.dl.util.concurrent.Callable;
import EDU.oswego.cs.dl.util.concurrent.Executor;
import EDU.oswego.cs.dl.util.concurrent.ThreadFactory;
import EDU.oswego.cs.dl.util.concurrent.ThreadFactoryUser;
import EDU.oswego.cs.dl.util.concurrent.PooledExecutor;
import EDU.oswego.cs.dl.util.concurrent.ThreadedExecutor;
import EDU.oswego.cs.dl.util.concurrent.DirectExecutor;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.Vector;

/**
 * ConnectionAdapter is an Adapter (see the GOF book) class that ties
 * together a Connection instance and a Thread to run it in. Java Purists
 * will immediately cringe at the use of the non-private member fields in
 * this class, and claim that they should be initialized via a constructor
 * instead of by direct manipulation (see run() in ConnectionManager). I
 * won't argue that they could be initialized that way. However, because
 * ConnectionAdapter isn't intended as a reusable component, but as an
 * integral part of the ConnectionManager component (it began life as an
 * anonymous class inside of ConnectionManager, that's how tightly these
 * two are tied together), and therefore subject to some relaxation of the
 * "normal" rules regarding encapsulation.
 *
 * One interesting trick it makes use of is the "Runnable finish;" field.
 * ConnectionManager sets this field to an anonymous Runnable class instance
 * that ConnectionAdapter must call before it shuts down completely; this
 * is what helps ConnectionManager keep track of the Connections still
 * outstanding. I could have, certainly, simply exposed the m_connections
 * Vector in ConnectionManager through methods like "add(Connection c)" and 
 * "remove(Connection c)", but even then there could be more than one
 * ConnectionManager running, so I would have had to have a ConnectionManager
 * reference in ConnectionAdapter, as well. This is more elegant, in my
 * opinion, and keeps the "cleanup" logic in the precise place where it
 * <bold>should</code> be--in the ConnectionManager class code.
 */
class ConnectionAdapter
    implements Runnable
{
    public void run()
    {
        try
        {
            connection.serve(socket, configInfo);
            
            socket.close();

            finish.run();
        }
        catch (InterruptedException intEx)
        {
            // Do nothing
        }
        catch (Exception ex)
        {
            ServerManager.log(ex);
        }
        finally
        {
            // Close the Socket
            try { socket.close(); }
            catch(java.io.IOException ioEx) { }

            ServerManager.log(
                getClass().getName() + ":" +
                "Socket closed");
        }
    }
    
    Connection connection;
    Socket socket;
    Runnable finish;
    ConfigProperties configInfo;
}


/**
 * ConnectionManager is a ThreadedServer that specifically manages a single
 * type of socket connection between this host and some anonymous client.
 *
 * <P><B>TODO:</B>
 * <UL>
 * <LI>Improve performance by object-pooling Connection instances
 * </UL>
 */
public class ConnectionManager extends SocketServer
{
    // Internal members
    //
    private Executor m_executor = null;
    private Vector m_connections = new Vector();
    private Class m_connectionClass = null;
    private static int s_count = 0;

    private ConfigProperty propThreads =
        new ConfigProperty("threads", new Integer(0),
                           "Maximum number of threads to use");
    private ConfigProperty propType =
        new ConfigProperty("type", new String(""),
                           "Connection class to use");
    private ConfigProperty propLoader =
        new ConfigProperty("loader", ClassLoaderStrategy.class,
                           null, null,
                           "ClassLoaderStrategy instance to use " +
                           "to load Connection instances",
                           null);
    private ConfigProperty propConnectionConfig =
        new ConfigProperty("connectionProperties",
                           ConfigProperties.class,
                           null, 
                           "ConfigProperties instance to use to " +
                           "hand to Connection instances",
                           null);

    private ConfigProperties m_configInfo =
        new ConfigProperties(super.getConfigInfo(), new ConfigProperty[]
    {
        propThreads,
        propType,
        propLoader,
        propConnectionConfig
    });
    
    
    /**
     * Start the Service.
     */
    public void start()
        throws Exception
    {
        setState(STARTING);
        
        ServerManager.log(
            "ConnectionManager: Loaded by " + 
            getClass().getClassLoader().getClass().getName());

        // Determine which Executor to use; unless the "threads"
        // ConfigProperty has been modified, we default to using
        // a ThreadedExecutor
        switch (((Integer)propThreads.getValue()).intValue())
        {
            case 0:
                m_executor = new ThreadedExecutor();
                break;
            case 1:
                m_executor = new DirectExecutor();
                break;
            default:
                Integer numThreads = 
                    (Integer)propThreads.getValue();
                m_executor = 
                    new PooledExecutor(numThreads.intValue());
                ((PooledExecutor)m_executor).waitWhenBlocked();
                break;
        }
        ServerManager.log(getClass().getName() + ": " +
            "Executor: " + m_executor.toString());

        // Determine Connection Class to use
        String name = (String)propType.getValue();
        if (name.equals(""))
        {
            throw new Exception("You must specify a \"type\" " +
                "argument to ConnectionManager");
        }
        else
        {
            ServerManager.log(
                getClass().getName() + ": Using " + 
                name + " as Connection type");
        
            // Determine if we've been given a specialized
            // ClassLoader instance to use; otherwise, just use
            // whatever ClassLoader loaded us
            ClassLoaderStrategy strat =
                (ClassLoaderStrategy)propLoader.getValue();
            if (strat == null)
            {
                ServerManager.log(
                    getClass().getName() + ": Using " + 
                    getClass().getClassLoader().getClass().getName()
                    + " as ClassLoader for " + name);
                m_connectionClass = Class.forName(name);
            }
            else
            {
                ServerManager.log(
                    getClass().getName() + ": Using " + 
                    strat.getClass().getName() + " as ClassLoader " +
                    "for " + name);
                
                StrategyClassLoader scl =
                    new StrategyClassLoader(strat);
                m_connectionClass = scl.loadClass(name);
            }
        }

        // Create a ThreadFactory that gives us a bit more
        // information in the Thread label; useful for tracing and
        // debugging. We only need to do this where we're using
        // Threads other than the 'main' Thread. (Both 
        // ThreadedExecutor and PooledExecutor are ThreadFactoryUser-
        // implementing classes.)
        if (m_executor instanceof ThreadFactoryUser)
        {
            ((ThreadFactoryUser)m_executor).setThreadFactory(new ThreadFactory()
            {
                public Thread newThread(Runnable cmd)
                {
                    return new Thread(cmd, 
                        m_connectionClass.getName() + s_count++);
                }
            });
        }
        
        // Call up the chain so SocketServer can do its thing
        super.start();
    }


    /**
     *
     */
    public ConfigProperties getConfigInfo()
    {
        // Retrieve ConfigProperties from Connection type, if a
        // getConfigInfo method is present
        //Class connectionClass = 
        return m_configInfo;
    }
    /**
     *
     */
    public void setConfigInfo(ConfigProperties props)
    {
        m_configInfo.set(props);
    }

    
    /**
     *
     */
    public void serve(Socket socket)
        throws Exception
    {
        // Create the Connection instance, place it in our list
        final Connection connection =
            (Connection)m_connectionClass.newInstance();
        m_connections.addElement(connection);
        
        // On each connection, create a new Runnable to be
        // executed within the Executor to do the actual work
        ConnectionAdapter ca = new ConnectionAdapter();
        ca.connection = connection;
        ca.socket = socket;

        ca.configInfo = new ConfigProperties(m_configInfo);
        ca.configInfo.remove("threads");
        ca.configInfo.remove("type");
        ca.configInfo.remove("loader");

        ca.finish = new Runnable()
        {
            public void run()
            {
                m_connections.removeElement(connection);
            }
        };
            // The "finish" Runnable in ConnectionAdapter
            // provides a Java-acceptable way of performing
            // a callback into this object to remove the
            // Connection object without having to (a) expose
            // the m_connections Vector to outside use via
            // "add" and/or "remove" methods, or (b) make
            // the m_connections Vector package-available
            // by reducing the 'private' access specifier to
            // '' (nothing, which is package access).
        
        m_executor.execute(ca);
    }
}    