package com.javageeks.gjas.services;

import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.net.Socket;
import java.util.Enumeration;
import java.util.Vector;
import com.javageeks.thread.PeriodicThread;
import com.javageeks.gjas.ConfigProperties;
import com.javageeks.gjas.ConfigProperty;
import com.javageeks.gjas.ServerManager;

/**
 * HeartbeatService sends a message to any listening clients every
 * <I>n</I> milliseconds. Put basically, this is the same publish-
 * subscribe behavior found in a variety of other places, such as
 * AWT/Swing's Event-EventListener system.
 *
 * This concept isn't necessarily limited solely to socket-based
 * communication--this could easily be adapted to other forms of
 * communicative technology, like RMI, CORBA, JMS, Mobile Objects,
 * and so on.
 */
public class HeartbeatService
    extends SocketServer
{
    // Internal members
    private Vector m_listeners = new Vector();
    private PeriodicThread m_pingThread;
    private ConfigProperty propInterval =
        new ConfigProperty("interval", new Integer(5 * 1000),
                           "Milliseconds between heartbeats");
    private ConfigProperty propMessage =
        new ConfigProperty("message", new String("PING"),
                           "Message to send on heartbeat");

    private ConfigProperties configInfo =
        new ConfigProperties(super.getConfigInfo(),
                             new ConfigProperty[]
    {
        propInterval,
        propMessage
    });
    
    
    /**
     * Inner class to store the Socket and associated streams
     */
    class Listener
    {
        public Listener(Socket socket)
            throws IOException
        {
            m_socket = socket;
            toSocket = new PrintWriter(
                new OutputStreamWriter(m_socket.getOutputStream()));
        }
    
        public Socket m_socket;
        public PrintWriter toSocket;
    }

    /**
     * Inner class to do the actual work of sending out the
     * "ping" messages
     */
    class Heartbeat
        implements Runnable
    {
        public void run()
        {
            // Synchronize on m_listeners to prevent anyone
            // from modifying the Vector while we're iterating
            // through it sending ping messages; it probably
            // wouldn't cause a major problem if it *did* happen,
            // since Vector protects against corruption within
            // itself, but 
            synchronized (m_listeners)
            {
                for (Enumeration enum = m_listeners.elements();
                     enum.hasMoreElements(); )
                {
                    // Get next element
                    Listener l = (Listener)enum.nextElement();
                    
                    // Send the ping message
                    l.toSocket.println(
                        (String)propMessage.getValue());
                    l.toSocket.flush();
                }
            }
        }
    }

    /**
     *
     */
    public HeartbeatService(int port, int interval, String pingMsg)
    {
        super(port);
        
        propInterval.setValue(new Integer(interval));
        propMessage.setValue(new String(pingMsg == null ? "PING" : pingMsg));
    }
    /**
     *
     */
    public HeartbeatService()
    { }
    

    /**
     *
     */
    public void start()
        throws Exception
    {
        // Call up to our base to allow SocketServer to do all its
        // work on our behalf
        super.start();

        // Start our ping thread
        m_pingThread = 
            new PeriodicThread(new Heartbeat(), 
                               ((Integer)propInterval.getValue()).intValue());
        m_pingThread.start();
    }
    /**
     *
     */
    public void stop()
        throws Exception
    {
        m_pingThread.interrupt();
        super.stop();
    }


    /**
     *
     */
    public ConfigProperties getConfigInfo()
    {
        return configInfo;
    }
    /**
     *
     */
    public void setConfigInfo(ConfigProperties props)
    {
        configInfo.set(props);

        // Check and reset interval value if different
        if (m_pingThread != null &&
            m_pingThread.getInterval() !=
            ((Integer)propInterval.getValue()).intValue())
        {
            m_pingThread.setInterval(((Integer)propInterval.getValue()).intValue());
        }
    }
    

    /**
     * Derived services must override this method. Once a client has
     * connected to us, this method is called to "do the work" of
     * handling the connection.
     */    
    public void serve(Socket socket)
        throws Exception
    {
        // Add this Socket to the list of Sockets we must broadcast
        // the "ping" message down
        m_listeners.addElement(new Listener(socket));
    }
}    