package com.javageeks.gjas.services;

import java.io.*;
import java.lang.reflect.*;
import java.net.*;
import java.util.*;


/**
 * RemoteObjectFactoryService
 */
public class RemoteObjectFactoryService
    extends com.javageeks.gjas.services.SocketServer
{
    /**
     * serve() takes a client request, sent via Serialization,
     * and extracts the class name, the constructor, and the
     * array of args to pass to the constructor, and proceeds
     * to attempt to construct an instance of that type. This
     * means that both the Service and the client must have the
     * bytecode of the exact Class returned available on the JVM's
     * CLASSPATH, or ClassNotFoundExceptions will result. (Note that
     * the client doesn't need to know exactly what type it's
     * getting back, it only needs to have it available via the
     * client's ClassLoader--a SocketClassLoader would effectively
     * make the entire mechanism load-on-the-fly.)
     */
    public void serve(Socket socket)
        throws Exception
    {
        // Set up
        ObjectOutputStream objOut = 
            new ObjectOutputStream(socket.getOutputStream());
        ObjectInputStream objIn = 
            new ObjectInputStream(socket.getInputStream());
            
        try
        {
            // Protocol: class-name (UTF), ctor rep (UTF), 
            // args (Vector)
            String className = objIn.readUTF();
            String ctorStringRep = objIn.readUTF();
            Vector args = (Vector)objIn.readObject();
            
            // First, get the Class object for className
            Class cls = Class.forName(className);
            
            // Next, find the Constructor corresponding to 
            // ctorStringRep
            Constructor[] allCtors = cls.getConstructors();
            Constructor ctor = null;
            for (int i=0; i<allCtors.length; i++)
            {
                if (allCtors[i].toString().equals(ctorStringRep))
                {
                    ctor = allCtors[i];
                    break;
                }
            }
            if (ctor == null)
                throw new Exception("Constructor " + ctorStringRep +
                    " not found on class " + cls.getName());
                    
            // Unpack the args into an Object array
            Object[] argsArray = new Object[args.size()];
            for (int i=0; i<argsArray.length; i++)
                argsArray[i] = args.elementAt(i);
                
            // Construct the instance
            Object instance = ctor.newInstance(argsArray);
            
            // Make sure instance is Serializable
            boolean foundIt = false;
            for (int j=0; j<cls.getInterfaces().length; j++)
            {
                String interfaceName = 
                    cls.getInterfaces()[j].getName();
                if (interfaceName.equals("java.io.Serializable"))
                {
                    foundIt = true;
                    break;
                }
            }
            if (!foundIt)
                throw new NotSerializableException(className);
                // Caught in the catch() block a few lines down and
                // sent back over the wire to the client
            
            // Send it over the wire
            objOut.writeUTF("SUCCESS");
            objOut.writeObject(instance);
        }
        catch (Exception ex)
        {
            objOut.writeUTF("ERROR");
            objOut.writeObject(ex);
        }
        
        objOut.flush();
    }
}    