package com.javageeks.gjas.services;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * RemoteStorageService
 */
public class RemoteStorageService
    extends com.javageeks.gjas.services.SocketServer
{
    /**
     * Use this Socket to answer client requests for serialization
     * services. This service, in contrast to others, services only
     * one request/response pair, then shuts down. This is
     * deliberate, as ObjectOutputStream and ObjectInputStream do
     * some activity to the target stream (the socket) during
     * their construction; this means that after the client makes
     * its request and quits, the server will get an Exception
     * claiming the socket "unexpectedly" quit.
     *
     * <P>Note also that order of construction of the streams (the
     * ObjectInput/ObjectOutput streams) <B>is</B> important;
     * because the client's ObjectInputStream is expecting to find
     * header information on the server's ObjectOutputStream, they
     * must be constructed in offsetting pairs. The server builds
     * them in Output-first-Input-second order, so the client must
     * build them in reverse (Input-first-Output-second) order. For
     * this reason, only use RemoteStorageRequest instances to
     * talk to the RemoteStorageService.
     */
    public void serve(Socket socket)
        throws Exception
    {
        // Set up
        ObjectOutputStream objOut = 
            new ObjectOutputStream(socket.getOutputStream());
        ObjectInputStream objIn = 
            new ObjectInputStream(socket.getInputStream());

        // Extract the command            
        String cmd = objIn.readUTF();
        if (cmd.equals("GET"))
        {
            // Protocol: GET, obj name
            String objName = (String)objIn.readObject();
            
            Object objInTable = m_objectTable.get(objName);
            if (objInTable != null)
            {
                objOut.writeUTF("SUCCESS");
                objOut.writeObject(objInTable);
            }
            else
            {
                objOut.writeUTF("ERROR");
                objOut.writeObject(
                    new Exception("Object not found"));
            }
            
            objOut.flush();
        } // GET
        else if (cmd.equals("CHECKIN"))
        {
            // Protocol: CHECKIN, client ID, obj name, obj
            String clientID = (String)objIn.readObject();
            String objName = (String)objIn.readObject();
            Object obj = objIn.readObject();

            // Is it there?
            Object objInTable = m_objectTable.get(objName);
            if (objInTable != null)
            {
                // It's there already; did we check it out?
                String ownerID = (String)m_checkOuts.get(objName);
                if (ownerID == null)
                {
                    // It's not checked out at all
                    objOut.writeUTF("ERROR");
                    objOut.writeObject(
                        new Exception("Object not checked out"));
                }
                else if (ownerID.equals(clientID))
                {
                    // Yes, 'tis checked out, and client owns it
                    
                    // Replace the old one; Hashtable allows dupes
                    m_objectTable.remove(objName);
                    m_objectTable.put(objName, obj);

                    // Remove the checkout
                    m_checkOuts.remove(objName);
                    
                    // Send success
                    objOut.writeUTF("SUCCESS");
                    objOut.writeObject(obj);
                }
                else
                {
                    // Yes, 'tis checked out, but client doesn't
                    // own it, so they can't check it back in
                    
                    objOut.writeUTF("ERROR");
                    objOut.writeObject(new Exception(
                        "Object checked out to " + ownerID));
                }
            }
            else
            {
                // It's not there; add it and return success
                m_objectTable.put(objName, obj);
                
                objOut.writeUTF("SUCCESS");
                objOut.writeObject(obj);
            }
            
            objOut.flush();
        } // CHECKIN
        else if (cmd.equals("CHECKOUT"))
        {
            // Protocol: CHECKOUT, client ID, obj name
            String clientID = (String)objIn.readObject();
            String objName = (String)objIn.readObject();
            
            // Is object there?
            Object objInTable = m_objectTable.get(objName);
            if (objInTable != null)
            {
                // Is it already checked out?
                String ownerID = (String)m_checkOuts.get(objName);
                if (ownerID == null)
                {
                    // No, nobody checked it out, so do so now
                    m_checkOuts.put(objName, clientID);
                    
                    objOut.writeUTF("SUCCESS");
                    objOut.writeObject(objInTable);
                        // We send back the object from the table
                        // because it's entirely possible that the
                        // object changed underneath the client from
                        // the last time they did a "GET"
                }
                else
                {
                    // Somebody (could be this client, it doesn't
                    // matter who) already checked it out
                    objOut.writeUTF("ERROR");
                    objOut.writeObject(new Exception(
                        "Object already checked out by " + 
                        ownerID));
                }
            }
            else
            {
                // Object's not in table; throw Exception
                objOut.writeUTF("ERROR");
                objOut.writeObject(new Exception(
                    "Object " + objName + " not found"));
            }

            objOut.flush();
        } // CHECKOUT
        else if (cmd.equals("DIFF"))
        {
            // Protocol: DIFF, objName, obj
            String objName = (String)objIn.readObject();
            Object obj = objIn.readObject();
            
            // Does object exist in table?
            Object objInTable = m_objectTable.get(objName);
            if (objInTable != null)
            {
                // Yes--compare and return
                if (objInTable.equals(obj))
                {
                    objOut.writeUTF("SUCCESS");
                    objOut.writeObject(new Boolean(true));
                }
                else
                {
                    objOut.writeUTF("SUCCESS");
                    objOut.writeObject(new Boolean(false));
                }
            }
            else
            {
                objOut.writeUTF("ERROR");
                objOut.writeObject(new Exception(
                    "Object " + objName + " not found"));
            }
            
            objOut.flush();
        } // DIFF
        else
        {
            objOut.writeUTF("ERROR");
            objOut.writeObject(
                new Exception("Unrecognized command: " + cmd));
        }
        
        objOut.flush();
    }
    
    
    // Internal members
    private Hashtable m_objectTable = new Hashtable();
    private Hashtable m_checkOuts = new Hashtable();
}
