package com.javageeks.net;

import java.io.*;

/**
 * This class serves the same purpose as the Java Socket class;
 * a simple abstraction of connecting to and using a named pipe.
 */
public class NamedPipe
{
    static
    {
        System.loadLibrary("NamedPipe");
    }

    // Constants
    //
    public static final int DUPLEX = 3;
    public static final int WRITE = 2;
    public static final int READ = 1;


    // Internal members
    //
    /*package-friendly*/ int m_hPipe;
    private int m_openMode;


    /**
     *
     */
    public NamedPipe()
    { }
    /**
     *
     */
    public NamedPipe(String pipeName, int openMode, int timeout)
        throws IOException
    {
        m_openMode = openMode;

        open(pipeName, openMode, timeout);
    }
    /**
     *
     */
    public void finalize()
    {
        close();
    }


    /**
     *
     */
    public void open(String pipeName, int openMode, int timeout)
        throws IOException
    {
        m_hPipe = nativeOpen(pipeName, openMode, timeout);
        if (m_hPipe == 0xFFFFFFFF)
        {
            throw new IOException("NT NamedPipe error");
        }
    }
    /**
     *
     */
    public void close()
    {
        nativeClose(m_hPipe);
    }


    /**
     *
     */
    public OutputStream getOutputStream()
        throws IOException
    {
        if ( (m_openMode == DUPLEX) ||
             (m_openMode == WRITE))
        {
            return new NamedPipeOutputStream(this);
        }
        else
            throw new IOException("Named-Pipe is inbound only");
    }
    /**
     *
     */
    public InputStream getInputStream()
        throws IOException
    {
        if ( (m_openMode == DUPLEX) ||
             (m_openMode == READ))
        {
            return new NamedPipeInputStream(this);
        }
        else
            throw new IOException("Named-Pipe is outbound only");
    }


    // Native methods
    //
    static private native
        int nativeOpen(String name, int mode, int timeOut);
    static private native    
        boolean nativeClose(int pipeHandle);


    // Test driver
    //
    public static void main (String args[])
        throws Exception
    {
        // Meant to run with MSVC NamedPipe server example
        // on the same machine; see the NT documentation on Win32
        // NamedPipes for more details on NamedPipes
        String pipe = "\\\\.\\pipe\\mynamedpipe";
        System.out.println("Attempting connection to " + pipe);

        // TEST One: One-shot write
        NamedPipe np1 =
            new NamedPipe(pipe, NamedPipe.DUPLEX, 20 * 1000);

        OutputStream npos1 = np1.getOutputStream();
        PrintWriter pw = new PrintWriter(npos1);
        pw.println("Generic Hello Message");
        pw.flush();

        // TEST Two: Back-and-forth read/write ("chat")
        NamedPipe np = 
            new NamedPipe(pipe, NamedPipe.DUPLEX, 20 * 1000);
        
        System.out.println("Connected");

        OutputStream npos = np.getOutputStream();
        InputStream npis = np.getInputStream();

        // Send a "Hello" message
        System.out.println("Sending message....");
        npos.write(pipe.getBytes());
        npos.flush();
        System.out.println("Done.");

        // Loop, sending and receiving messages from the server
    }
}
