import java.io.*;
import java.net.*;
import java.util.*;

/**
 * RemoteStorageClient
 */
public class RemoteStorageClient
    implements Serializable
{
    private RemoteStorageClient()
    {
        // Can't instantiate a default no-arg instance of this guy
    }

    /**
     * Constructor.
     */
    public RemoteStorageClient(String ID, String host, int port)
    {
        m_clientID = ID;
        m_host = host;
        m_port = port;
    }

    /**
     * Retrieves a "read-only" (that is, you don't own the lock on
     * this object) object by name.
     */    
    public Object get(String objName)
        throws Exception
    {
        Socket socket = new Socket(m_host, m_port);
        ObjectInputStream fromSocket = 
            new ObjectInputStream(socket.getInputStream());
        ObjectOutputStream toSocket = 
            new ObjectOutputStream(socket.getOutputStream());

        // Send the request: "GET", name
        toSocket.writeUTF("GET");
        toSocket.writeObject(objName);
        toSocket.flush();

        // Check the repsonse; if the response string is anything
        // other than "SUCCESS", throw the next object pulled from
        // the stream (which RemoteStorageService guarantees will be
        // an Exception type)
        String response = fromSocket.readUTF();
        if (response.equals("SUCCESS"))
        {
            return fromSocket.readObject();
        }
        else
        {
            throw (Exception)fromSocket.readObject();
        }
    }
    
    /**
     * Retrieves the object by name and locks it for exclusive
     * modification by this client.
     */
    public Object checkOut(String objName)
        throws Exception
    {
        Socket socket = new Socket(m_host, m_port);
        ObjectInputStream fromSocket = 
            new ObjectInputStream(socket.getInputStream());
        ObjectOutputStream toSocket = 
            new ObjectOutputStream(socket.getOutputStream());

        // Send the request: "CHECKOUT", client ID, obj name
        toSocket.writeUTF("CHECKOUT");
        toSocket.writeObject(m_clientID);
        toSocket.writeObject(objName);
        toSocket.flush();

        // Check the repsonse; if the response string is anything
        // other than "SUCCESS", throw the next object pulled from
        // the stream (which RemoteStorageService guarantees will be
        // an Exception type)
        String response = fromSocket.readUTF();
        if (response.equals("SUCCESS"))
        {
            return fromSocket.readObject();
        }
        else
        {
            throw (Exception)fromSocket.readObject();
        }
    }
    
    /**
     * Stores the object <code>obj</code> under the name
     * <code>objName</code>, but only if the object is either
     * checked out to this client, or not already in the system.
     */
    public Object checkIn(String objName, Object obj)
        throws Exception
    {
        Socket socket = new Socket(m_host, m_port);
        ObjectInputStream fromSocket = 
            new ObjectInputStream(socket.getInputStream());
        ObjectOutputStream toSocket = 
            new ObjectOutputStream(socket.getOutputStream());

        // Send the request: "CHECKIN", ID, name, obj
        toSocket.writeUTF("CHECKIN");
        toSocket.writeObject(m_clientID);
        toSocket.writeObject(objName);
        toSocket.writeObject(obj);
        toSocket.flush();

        // Check the repsonse; if the response string is anything
        // other than "SUCCESS", throw the next object pulled from
        // the stream (which RemoteStorageService guarantees will be
        // an Exception type)
        String response = fromSocket.readUTF();
        if (response.equals("SUCCESS"))
        {
            return fromSocket.readObject();
        }
        else
        {
            throw (Exception)fromSocket.readObject();
        }
    }

    /**
     * Compare the object stored in the system under the name
     * <code>objName</code> to the object passed in <code>obj</code>
     */
    public boolean diff(String objName, Object obj)
        throws Exception
    {
        Socket socket = new Socket(m_host, m_port);
        ObjectInputStream fromSocket = 
            new ObjectInputStream(socket.getInputStream());
        ObjectOutputStream toSocket = 
            new ObjectOutputStream(socket.getOutputStream());

        // Send the request: "DIFF", name, obj
        toSocket.writeUTF("DIFF");
        toSocket.writeObject(objName);
        toSocket.writeObject(obj);
        toSocket.flush();

        // Check the repsonse; if the response string is anything
        // other than "SUCCESS", throw the next object pulled from
        // the stream (which RemoteStorageService guarantees will be
        // an Exception type)
        String response = fromSocket.readUTF();
        if (response.equals("SUCCESS"))
        {
            // Pull back the Boolean
            return ((Boolean)fromSocket.readObject()).booleanValue();
        }
        else
        {
            throw (Exception)fromSocket.readObject();
        }
    }


    // Internal members
    private String m_clientID;
    private String m_host;
    private int m_port;


    /**
     * Test driver
     */
    public static void main(String[] args)
        throws Exception
    {
        if (args.length < 1)
        {
            System.out.println(
                "Usage: java RemoteStorageClient <hostname:port>");
            return;
        }
        
        // Parse out hostname and port
        String host;
        Integer port;
        host = args[0].substring(0, args[0].indexOf(":"));
        port = new Integer(args[0].substring(
            args[0].indexOf(":")+1, args[0].length()));
        
        System.out.println("Connecting to " + host + ":" + port);
        
        RemoteStorageClient client = 
            new RemoteStorageClient("Test", host, port.intValue());

        // Add objects to the storage database
        try
        {
            client.checkIn("Test Data", 
                new String("This is a test"));
            client.checkIn("Today's Date", new Date());
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
            System.out.println("If you didn't restart the server " +
                "after the last test run, this is normal.");
        }
        
        // Retrieve an object that's there
        String testData = (String)client.get("Test Data");
        System.out.println(
            "client.get(\"Test Data\") = " + testData);
        
        // Retrieve an object that's NOT there
        try
        {
            client.get("Not there");
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
            System.out.println("This is normal; object shouldn't be there");
        }
        
        // Check out an object as "Test"
        Object obj = client.checkOut("Today's Date");
        
        // Create a new client named "Other"
        RemoteStorageClient otherClient = 
            new RemoteStorageClient("Other", host, port.intValue());

        // Try to check out same object "Test" checked out
        try
        {
            Object otherObj = otherClient.checkOut("Today's Date");
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
            System.out.println("Two clients shouldn't be able to " +
                "check the same object out, so this is normal.");
        }
        
        // Try to check in object "Test" checked out
        try
        {
            otherClient.checkIn("Today's Date", new Date());
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
            System.out.println("Only the client that checked out" +
                " an object should be allowed to check it back in," +
                " so this is normal.");
        }
        
        // Get object as "Other"; should get back old value
        Object oldValue = otherClient.get("Today's Date");
        
        // Have "Test" check in object
        Date date = (Date)oldValue;
        System.out.println(date);
        Date newDate = new Date();
        System.out.println("Changing to " + newDate);
        client.checkIn("Today's Date", newDate);
        
        // Run diff between "Other"'s retrieved object and what
        // the RemoteStorageService now says is the object
        boolean diff = otherClient.diff("Today's Date", oldValue);
        if (diff == false)
            System.out.println(newDate + " != " + oldValue);
        else
            System.out.println(newDate + " == " + oldValue);
    }
}    