/**
 * PeriodicThread is a specific type of Thread that fires off its
 * associated Runnable evry <code>interval</code> milliseconds.
 */
public class PeriodicThread extends Thread
{
    private PeriodicThread()
    {
        // This prevents instantiation without an associated Runnable;
        // I don't want to allow the possibility of this code from
        // ever compiling:
        //
        // new PeriodicThread().start()
        //
    }

    /**
     * Constructor taking the Runnable whose <code>run</code> method
     * we wish to call every <code>interval</code> milliseconds.
     */
    public PeriodicThread(Runnable r, int interval)
    {
        super();
        
        m_runnable = r;
        m_interval = interval;
    }

    /**
     * The <code>run</code> method spins in an infinite loop, calling
     * run on the owned Runnable instance every <code>interval</code>
     * milliseconds (as specified in the constructor). The time spent
     * in the Runnable's <code>run</code> method is not taken into
     * account in the period spent sleeping.
     */
    public void run()
    {
        try
        {
            while (true)
            {
                Thread.sleep(m_interval);
                m_runnable.run();
            }
        }
        catch (InterruptedException iEx)
        {
            return;
        }
    }
    
    private int m_interval = 0;
    private Runnable m_runnable = null;
    
    
    /**
     * Test driver for the PeriodicThread component
     */
    public static void main(String[] args)
        throws Exception
    {
        PeriodicThread pt = new PeriodicThread(new Runnable() 
        {
            public void run()
            {
                System.out.println("Fired!");
            }
        }, 10 * 1000);
        pt.start();
        
        PeriodicThread pt2 = new PeriodicThread(new Runnable()
        {
            public void run()
            {
                System.out.println("Hired!");
            }
        }, 15*1000);
        pt2.start();
        
        System.out.println("Use Ctrl-C to quit.");
        pt.join();
        pt2.join();
    }
}